
from typing import Dict
from pyot.pipeline.token import PipelineToken

from .base import BaseEndpoint


class CDragonEndpoint(BaseEndpoint):

    base_url = "https://raw.communitydragon.org"

    all = {
        "lol": {
            "cdragon_champion_by_id": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/champions/{id}.json",
            "cdragon_champion_summary": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/champion-summary.json",
            "cdragon_item_full": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/items.json",
            "cdragon_rune_full": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/perks.json",
            "cdragon_spells_full": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/summoner-spells.json",
            "cdragon_profile_icon_full": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/profile-icons.json",
        },
        "tft": {
            "cdragon_tft_full": "/{version}/cdragon/tft/{locale}.json",
            "cdragon_profile_icon_full": "/{version}/plugins/rcp-be-lol-game-data/global/{locale}/v1/profile-icons.json",
        }
    }

    def clean(self, token: PipelineToken) -> Dict[str, str]:
        if token.model == "lol" and token.params["locale"].lower() == "en_us":
            token.params["locale"] = "default"
        elif token.model == "tft":
            if token.method in self.all["lol"] and token.params["locale"].lower() == "en_us":
                token.params["locale"] = "default"
            elif token.params["locale"].lower() == "default":
                token.params["locale"] = "en_us"
        return token.params




from asyncio import iscoroutinefunction
from inspect import isasyncgenfunction
from functools import cached_property as untyped_cached_property
from typing import Any, AsyncGenerator, Awaitable, Callable, Generic, TypeVar


R = TypeVar("R")

IY = TypeVar("IY")
IS = TypeVar("IS")


class async_property(Generic[R]):
    '''
    Async equivalent of `property`, takes an async method and
    converts it to a property that returns an awaitable with the return value.
    
    Usage:
    ```python
    class A:
        @async_property
        async def b(self):
            ...
    a = A()
    await a.b
    ```
    '''
    name = None

    @classmethod
    def func(cls, instance: Any) -> Any:
        raise TypeError(
            f'Cannot use {cls.__class__.__name__} instance without calling '
            '__set_name__() on it.'
        )

    def __init__(self, func: Callable[..., Awaitable[R]], name=None):
        assert iscoroutinefunction(func), "Function is not async"
        self.real_func = func
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name
            self.func = self.real_func
        elif name != self.name:
            raise TypeError(
                f"Cannot assign the same {self.__class__.__name__} to two different names "
                f"({self.name} and {name})."
            )

    def __get__(self, instance, cls=None) -> Awaitable[R]:
        if instance is None:
            return self
        return self.proxy(instance)

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")

    async def proxy(self, instance: Any) -> Awaitable[R]:
        res = await self.func(instance)
        return res


class async_cached_property(async_property, Generic[R]):
    '''
    Async equivalent of `functools.cached_property`, takes an async method and
    converts it to a cached property that returns an awaitable with the return value.
    
    Usage:
    ```python
    class A:
        @async_cached_property
        async def b(self):
            ...
    a = A()
    await a.b
    ```
    '''
    name = None

    def __init__(self, func: Callable[..., Awaitable[R]], name=None):
        assert iscoroutinefunction(func), "Function is not async"
        self.real_func = func
        self.once = False
        self.__doc__ = getattr(func, '__doc__')

    async def proxy(self, instance: Any) -> Awaitable[R]:
        if self.once:
            try:
                return instance.__dict__[self.name]
            except KeyError:
                pass
        self.once = True
        res = instance.__dict__[self.name] = await self.func(instance)
        return res


class async_generator_property(async_property, Generic[IY, IS]):
    '''
    Modified version of `async_property`, intended for use in async generators.
    The return typing of the decorated property is: AsyncGenerator[...]
    
    Usage:
    ```python
    class A:
        @async_generator_property
        async def b(self):
            yield ...
    a = A()
    async for _ in a.b:
        ...
    ```
    '''

    def __init__(self, func: Callable[..., AsyncGenerator[IY, IS]], name=None):
        assert isasyncgenfunction(func), "Function is not async generator"
        self.real_func = func
        self.__doc__ = getattr(func, '__doc__')

    def __get__(self, instance, cls=None) -> Awaitable[R]:
        if instance is None:
            return self
        return self.proxy(instance)

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")

    async def proxy(self, instance: Any) -> Awaitable[R]:
        async for iterating in self.func(instance):
            yield iterating


def cached_property(func: Callable[..., R]) -> R:
    untyped_cached_property(func)



from typing import List, Iterator, TYPE_CHECKING
from datetime import datetime, timedelta

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .summoner import Summoner
    from .champion import Champion
    from .item import Item
    from .trait import Trait


# PYOT STATIC OBJECTS

class MatchMetadataData(PyotStatic):
    id: str
    data_version: str
    participant_puuids: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"participant_puuids"}
        renamed = {"match_id": "id", "participants": "participant_puuids"}

    @property
    def participants(self) -> List["Summoner"]:
        from .summoner import Summoner
        return [Summoner(puuid=i, platform=self.id.split('_')[0]) for i in self.participant_puuids]


class MatchInfoCompanionData(PyotStatic):
    content_id: str
    skin_id: int
    species: str


class MatchInfoTraitData(PyotStatic):
    name: str
    num_units: int
    style: int
    tier_current: int
    tier_total: int

    @property
    def trait(self) -> "Trait":
        from .trait import Trait
        return Trait(key=self.name)


class MatchInfoUnitData(PyotStatic):
    item_ids: List[int]
    champion_key: str
    chosen: str
    name: str
    rarity: int
    tier: int

    class Meta(PyotStatic.Meta):
        raws = {"item_ids"}
        renamed = {"items": "item_ids", "character_id": "champion_key"}

    @property
    def items(self) -> List["Item"]:
        from .item import Item
        return [Item(id=i) for i in self.item_ids]

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(key=self.champion_key)


class MatchInfoParticipantData(PyotStatic):
    companion: MatchInfoCompanionData
    gold_left: int
    last_round: int
    level: int
    placement: int
    players_eliminated: int
    puuid: str
    time_eliminated_secs: float
    total_damage_to_players: int
    traits: List[MatchInfoTraitData]
    units: List[MatchInfoUnitData]
    _pyot_calculated_platform: str

    class Meta(PyotStatic.Meta):
        renamed = {"time_eliminated": "time_eliminated_secs"}

    @property
    def time_eliminated(self) -> timedelta:
        return timedelta(seconds=self.time_eliminated_secs)

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(puuid=self.puuid, platform=self._pyot_calculated_platform)


class MatchInfoData(PyotStatic):
    datetime_millis: int
    length_secs: float
    variation: str
    version: str
    participants: List[MatchInfoParticipantData]
    queue_id: int
    tft_mode: str
    tft_set_number: int

    class Meta(PyotStatic.Meta):
        renamed = {"game_datetime": "datetime_millis", "game_length": "length_secs", "game_variation": "variation", "game_version": "version"}

    @property
    def datetime(self) -> datetime:
        return datetime.fromtimestamp(self.datetime_millis//1000)

    @property
    def length(self) -> timedelta:
        return timedelta(seconds=self.length_secs)


# PYOT CORE OBJECTS

class Match(PyotCore):
    id: str
    info: MatchInfoData
    metadata: MatchMetadataData

    class Meta(PyotCore.Meta):
        rules = {"match_v1_match": ["id"]}

    def __init__(self, id: str = empty, region: str = models.tft.DEFAULT_REGION):
        self.initialize(locals())

    def transform(self, data):
        platform = self.id.split("_")[0]
        for i in range(len(data["info"]["participants"])):
            data["info"]["participants"][i]["_pyot_calculated_platform"] = platform
        return data


class MatchHistory(PyotCore):
    ids: List[str]
    puuid: str

    class Meta(PyotCore.Meta):
        rules = {"match_v1_matchlist": ["puuid"]}
        raws = {"ids"}

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return Match(id=self.ids[item], region=self.region)

    def __iter__(self) -> Iterator[Match]:
        return iter(self.matches)

    def __len__(self):
        return len(self.ids)

    def __init__(self, puuid: str = empty, region: str = models.tft.DEFAULT_REGION):
        self.initialize(locals())

    def query(self, count: int = 20):
        '''Set query request parameters.'''
        super()._place_query(locals())
        return self

    @property
    def matches(self) -> List[Match]:
        return [Match(id=id_, region=self.region) for id_ in self.ids]

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        try:
            return Summoner(account_id=self.account_id, platform=self.ids[0].split("_")[0])
        except (AttributeError, IndexError) as e:
            raise AttributeError("Match history is empty, could not identify platform from list") from e

    def transform(self, data):
        return {"ids": data}



from importlib import import_module
from typing import Dict, List, Tuple, Type, Union
from types import FunctionType, ModuleType
import inspect

from pyot.core.objects import PyotStaticBase


def get_forward_refs(model_name: str) -> Dict[str, Union[PyotStaticBase, ModuleType]]:
    classes = {}

    model = import_module(f"pyot.models.{model_name}")
    for attr in list(filter(lambda att: not att.startswith("__"), dir(model))):
        classes[attr] = getattr(model, attr)
    return classes


def get_module_locals(o: Type):
    classes = {}

    module = inspect.getmodule(o)
    for attr in list(filter(lambda attr: not attr.startswith("__"), dir(module))):
        classes[attr] = getattr(module, attr)
    return classes


def get_properties(o: PyotStaticBase) -> List[Tuple[str, FunctionType]]:
    return inspect.getmembers(o.__class__, predicate=lambda func: isinstance(func, property))



from typing import List, Iterator, Dict, Union

from pyot.conf.model import models
from pyot.core.functional import cache_indexes, lazy_property, empty
from pyot.core.exceptions import NotFound
from pyot.utils.tft.cdragon import abs_url, merge_set_data
from pyot.utils.lol.cdragon import sanitize_description
from .base import PyotCore, PyotStatic


# PYOT STATIC OBJECT

class TraitEffectData(PyotStatic):
    max_units: int
    min_units: int
    style: int
    variables: Dict[str, Union[float, str]]

    class Meta(PyotStatic.Meta):
        raws = {"variables"}


# PYOT CORE OBJECT

class Trait(PyotCore):
    set: int
    key: str
    name: str
    effects: List[TraitEffectData]
    icon_path: str
    description: str

    class Meta(PyotCore.Meta):
        rules = {"cdragon_tft_full": ["?set", "?key", "version", "locale"]}
        renamed = {"api_name": "key", "desc": "description", "icon": "icon_path"}

    def __init__(self, key: str = empty, set: int = empty, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())
        if key and set is None:
            self.find_set()

    def find_set(self):
        try:
            self.set = int(self.key.split("_")[0][3:])
        except Exception:
            self.set = -1

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(self.key, merge_set_data(data, self.set, "traits"), "apiName")

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @lazy_property
    def cleaned_description(self) -> str:
        return sanitize_description(self.description)


class Traits(PyotCore):
    set: int
    traits: List[Trait]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_tft_full": ["?set", "version", "locale"]}

    def __init__(self, set: int = -1, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.traits[item]

    def __iter__(self) -> Iterator[Trait]:
        return iter(self.traits)

    def __len__(self):
        return len(self.traits)

    def filter(self, data):
        try:
            return merge_set_data(data, self.set, "traits")
        except KeyError as e:
            raise NotFound("Request was successful but filtering gave no matching item") from e

    def transform(self, data):
        return {"traits": data}



from typing import TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore

if TYPE_CHECKING:
    from .summoner import Summoner


class ThirdPartyCode(PyotCore):
    code: str
    summoner_id: str

    class Meta(PyotCore.Meta):
        rules = {"third_party_code_v4_code": ["summoner_id"]}

    def __init__(self, summoner_id: str = empty, platform: str = models.tft.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        new_data = {}
        new_data["code"] = data
        return new_data

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)



from typing import Any, Dict, Union
from functools import partial, wraps

from pyot.utils.text import camel_case
from .exceptions import NotFound


class lazy_property(property):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance and inserted into the meta data dict using camel_case key.\n
    Can only be used on pyot class methods.
    """
    name = None

    @classmethod
    def func(cls, instance) -> Any:
        raise TypeError(
            f'Cannot use {cls.__class__.__name__} instance without calling '
            '__set_name__() on it.'
        )

    def __init__(self, func, name=None): # pylint: disable=super-init-not-called
        self.real_func = func
        self.key = camel_case(func.__name__)
        self.once = False
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        from .objects import PyotStaticBase
        if not issubclass(owner, PyotStaticBase):
            raise TypeError('Cannot use lazy_property on non pyot class methods.')
        if self.name is None:
            self.name = name
            self.func = self.real_func
        elif name != self.name:
            raise TypeError(
                "Cannot assign the same lazy_property to two different names "
                "(%r and %r)." % (self.name, name)
            )

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        if self.once:
            try:
                return instance._meta.data[self.key]
            except KeyError:
                pass
        self.once = True
        res = instance._meta.data[self.key] = instance.__dict__[self.name] = self.func(instance)
        return res

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")


class cache_indexes:

    def __init__(self, func):
        self.indexes = {}
        self.func = func

    def __get__(self, instance, cls=None):
        return partial(self.__call__, instance)

    def __call__(self, instance, data):
        try:
            return self.func(instance, self, data)
        except KeyError as e:
            raise NotFound("Request was successful but filtering gave no matching item") from e

    @staticmethod
    def _get_index(name, data, key):
        for ind, item in enumerate(data):
            if item[key] == name:
                return ind
        raise NotFound("Request was successful but filtering gave no matching item")

    def get(self, name: str, data: Dict, key: Union[str, int]):
        lazy_func = partial(self._get_index, name, data, key)
        ind = self._get(name, lazy_func)
        if data[ind][key] == name: # RETURN ONLY IF ID MATCHES
            return data[ind]
        ind = lazy_func()
        self._set(name, ind)
        return data[ind]

    def _get(self, name: str, func):
        '''Get an index'''
        try:
            return self.indexes[name]
        except KeyError:
            pass
        response = func()
        self._set(name, response)
        return response

    def _set(self, name: str, val):
        '''Set an index.'''
        self.indexes[name] = val
        return name

    def _clear(self):
        '''Clear indexes.'''
        self.indexes = dict()


def turbo_copy(data, top, level=0):
    if level >= top:
        return data
    if isinstance(data, dict):
        data = data.copy()
        for key, val in data.items():
            data[key] = turbo_copy(val, top, level + 1)
    elif isinstance(data, list):
        data = data.copy()
        for key, val in enumerate(data):
            data[key] = turbo_copy(val, top, level + 1)
    return data


def is_laziable(obj):
    if isinstance(obj, dict) or isinstance(obj, list):
        return True
    return False


def convert_keys_camel_case(kwargs: Dict) -> Dict:
    return {camel_case(key): val for (key, val) in kwargs.items() if val is not None}


empty = None



from typing import Dict, List, Mapping, Any, Set, Tuple, Type, Union, get_type_hints
import inspect
import re

from pyot.conf.pipeline import pipelines
from pyot.pipeline.core import Pipeline
from pyot.pipeline.token import PipelineToken
from pyot.utils.copy import fast_copy

from .functional import lazy_property, is_laziable, convert_keys_camel_case


class PyotLazy:

    def __init__(self, container: Union[None, Type[list], Type[dict]], clas: Union[Type["PyotStaticBase"], Type["PyotCoreBase"]], obj: Any, root: "PyotCoreBase"):
        self.container = container
        self.clas = clas
        self.root = root
        self.obj = obj

    def __call__(self):
        try:
            if issubclass(self.clas, PyotCoreBase):
                if self.container is None:
                    return self.load_core(self.obj)
                if issubclass(self.container, list):
                    return [self.load_core(obj) for obj in self.obj]
                if issubclass(self.container, dict):
                    return {okey: self.load_core(obj) for okey, obj in self.obj.items()}
                raise TypeError(f"Invalid container type '{self.container.__name__}'")
            if self.container is None:
                return self.load_static(self.obj)
            if issubclass(self.container, list):
                return [self.load_static(obj) for obj in self.obj]
            if issubclass(self.container, dict):
                return {okey: self.load_static(obj) for okey, obj in self.obj.items()}
            raise TypeError(f"Invalid container type '{self.container.__name__}'")
        except Exception as e:
            raise RuntimeError(f"Failed to lazy load '{self.clas.__name__}' object due to: ({type(e)}) {e}") from e

    def load_static(self, obj):
        instance: "PyotStaticBase" = self.clas(obj)
        instance._meta.root = self.root
        return instance.fill()

    def load_core(self, obj):
        kwargs = {}
        if "version" in self.clas.Meta.arg_names:
            try: kwargs["version"] = self.root.version
            except AttributeError: pass
        if "locale" in self.clas.Meta.arg_names:
            try: kwargs["locale"] = self.root.locale
            except AttributeError: pass
        if "platform" in self.clas.Meta.arg_names and "platform" in self.root.Meta.arg_names:
            try: kwargs["platform"] = self.root.platform
            except AttributeError: pass
        if "region" in self.clas.Meta.arg_names and "region" in self.root.Meta.arg_names:
            try: kwargs["region"] = self.root.region
            except AttributeError: pass
        instance: "PyotCoreBase" = self.clas(**kwargs)
        instance._meta.root = self.root
        instance._meta.data = instance.transform(obj)
        return instance.fill()


class PyotRoutingBase:

    class Meta:
        root: "PyotCoreBase"
        pipeline: Pipeline

    _meta: Meta
    _region: str = None
    _platform: str = None
    _locale: str = None
    _version: str = None
    _regions: Set[str] = set()
    _platforms: Set[str] = set()
    _platform2regions: Dict[str, str] = {None: None}

    @property
    def region(self) -> str:
        val = self._region or self._meta.root._region or \
            self._platform2regions.get(self._platform or self._meta.root._platform, None)
        if val is None:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'region'")
        return val

    @region.setter
    def region(self, val: str):
        val = val.lower()
        if val not in self._regions:
            raise ValueError(f"'{val}' is not a valid region")
        self._region = val

    @property
    def platform(self) -> str:
        val = self._platform or self._meta.root._platform
        if val is None:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'platform'")
        return val

    @platform.setter
    def platform(self, val: str):
        val = val.lower()
        if val not in self._platforms:
            raise ValueError(f"'{val}' is not a valid platform")
        self._platform = val

    @property
    def locale(self) -> str:
        val = self._locale or self._meta.root._locale
        if val is None:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'locale'")
        return val

    @locale.setter
    def locale(self, val: str):
        self._locale = val

    @property
    def version(self) -> str:
        val = self._version or self._meta.root._version
        if val is None:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'version'")
        return val

    @version.setter
    def version(self, val: str):
        self._version = val

    @property
    def metaroot(self) -> "PyotCoreBase":
        return self._meta.root

    @property
    def metapipeline(self) -> Pipeline:
        return self._meta.pipeline


class PyotMetaClass(type):

    def __new__(cls, name, bases, attrs):
        if 'Meta' not in attrs and cls.is_static_core(bases):
            attrs['Meta'] = type('Meta', (cls.get_static_core(bases).Meta,), {'__module__': attrs['__module__'] + f".{name}"})
        clas: "PyotStaticBase" = super().__new__(cls, name, bases, attrs)
        clas.Meta.types = cls.get_types(clas)
        clas.Meta.nomcltrs = {}
        if cls.is_static_core(bases):
            if issubclass(clas, PyotCoreBase):
                try:
                    arg_names = set(inspect.getfullargspec(clas.__init__).args)
                    cls.set_server_type(clas, arg_names)
                    clas.Meta.arg_names = arg_names
                except TypeError:
                    pass
            clas.Meta.lazy_props = [prop.split(".") for prop in cls.get_lazy_props(clas, [])]
        return clas

    @staticmethod
    def is_static_core(bases):
        base_names = set()
        for base in bases:
            base_names |= {cl.__name__ for cl in inspect.getmro(base)}
        return 'PyotStatic' in base_names or 'PyotCore' in base_names

    @staticmethod
    def get_static_core(bases):
        deep_bases = set()
        for base in bases:
            deep_bases |= set(inspect.getmro(base))
        try:
            return next(base for base in deep_bases if base.__name__ == 'PyotCore')
        except StopIteration:
            return next(base for base in deep_bases if base.__name__ == 'PyotStatic')

    @staticmethod
    def get_types(clas: "PyotStaticBase"):
        types = get_type_hints(clas)
        for typ, clas in types.items():
            try:
                if clas.__origin__ is list:
                    types[typ] = (clas.__args__[0], list)
                elif clas.__origin__ is dict:
                    types[typ] = (clas.__args__[1], dict)
                else:
                    types[typ] = (clas, None)
            except Exception:
                types[typ] = (clas, None)
        return types

    @staticmethod
    def get_lazy_props(clas: "PyotStaticBase", props, prefix=""):
        props += [prefix + p for p in dir(clas) if isinstance(getattr(clas, p), lazy_property)]
        types = {attr:cl for attr, cl in clas.Meta.types.items() if inspect.isclass(cl) and issubclass(cl, PyotStaticBase)}
        for typ, cl in types.items():
            PyotMetaClass.get_lazy_props(cl, props, prefix + typ + ".")
        return props

    @staticmethod
    def set_server_type(clas: "PyotCoreBase", args):
        for server in clas.Meta.server_type_names:
            if server in args:
                clas.Meta.server_type = server
                return


class PyotStaticBase(PyotRoutingBase, metaclass=PyotMetaClass):

    class Meta(PyotRoutingBase.Meta):
        # Mutable objects should be overriden on inheritance
        server_type: str = None
        lazy_props: List[str]
        nomcltrs: Dict[str, Any]
        types: Dict[str, Tuple[Any, Union[None, Type[list], Type[dict]]]]
        data: Dict[str, Any]
        raws: Set[str] = set()
        renamed: Dict[str, str] = {}

    _meta: Meta

    def __init__(self, data):
        # Instantiate Meta class, isolating data dict
        self._meta = self.Meta()
        self._meta.data = data

    def __getattr__(self, name):
        try:
            lazy = self.__dict__.pop('_lazy__' + name)
            obj = lazy()
            setattr(self, name, obj)
            return obj
        except (KeyError, AttributeError) as e:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'") from e

    def __getitem__(self, item):
        return self._meta.data[item]

    def qualkey(self, key: str) -> str:
        name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', key)
        newkey = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()
        if newkey in self._meta.renamed:
            newkey = self._meta.renamed[newkey]
        return newkey

    def fill(self):
        mapping = self._meta.nomcltrs

        for key, val in self._meta.data.items():
            try:
                attr = mapping[key]
            except KeyError:
                attr = self.qualkey(key)
                mapping[key] = attr

            if is_laziable(val):
                if attr in self._meta.raws:
                    setattr(self, attr, val)
                    continue
                try:
                    attr_type = self._meta.types[attr]
                    setattr(self, '_lazy__' + attr, PyotLazy(attr_type[1], attr_type[0], val, self._meta.root))
                except KeyError:
                    setattr(self, attr, val)
            else:
                setattr(self, attr, val)

        return self

    def load_lazy_properties(self, instance, prop, ind=0):
        if ind == len(prop):
            return
        try:
            attr = getattr(instance, prop[ind])
        except AttributeError:
            return
        if isinstance(attr, list):
            for val in attr:
                self.load_lazy_properties(val, prop, ind + 1)
        elif isinstance(attr, dict):
            for val in attr.values():
                self.load_lazy_properties(val, prop, ind + 1)
        else:
            self.load_lazy_properties(attr, prop, ind + 1)

    def rdict(self):
        dic = {}
        for key, (val, container) in self._meta.types.items():
            if key.startswith("_"):
                continue
            try:
                obj = getattr(self, key)
            except AttributeError:
                continue
            try:
                if issubclass(val, PyotStaticBase):
                    try:
                        if container is None:
                            dic[key] = obj.rdict()
                        elif issubclass(container, list):
                            dic[key] = [ob.rdict() for ob in obj]
                        elif issubclass(container, dict):
                            dic[key] = {okey: ob.rdict() for okey, ob in obj.items()}
                        else:
                            raise TypeError(f"Invalid container type '{container.__name__}'")
                    except AttributeError:
                        pass
                else:
                    dic[key] = obj
            except TypeError:
                pass
        return dic

    def dict(self, force_copy=False, lazy_props=False, recursive=False):
        if lazy_props:
            for prop in self._meta.lazy_props:
                self.load_lazy_properties(self, prop)
        dic = self.rdict() if recursive else self._meta.data
        return fast_copy(dic) if force_copy else dic.copy()


class PyotCoreBase(PyotStaticBase):

    class Meta(PyotStaticBase.Meta):
        # Mutable objects should be overriden on inheritance
        key: str
        server: str
        load: Mapping[str, Any]
        query: Mapping[str, Any]
        body: Mapping[str, Any]
        arg_names: Set[str]
        server_type: str
        server_type_names = {"platform", "region"}
        rules: Mapping[str, List[str]] = {}
        raw_data: Any

    _meta: Meta

    def initialize(self, kwargs: Dict):
        # Instantiate meta class and fill kwargs
        self._meta = self.Meta()
        self._meta.root = self
        self._meta.query = {}
        self._meta.data = {}
        self._meta.body = {}

        kwargs.pop("self", None)
        kwargs.pop("__class__", None)
        for name, val in kwargs.items():
            if val is not None:
                self._meta.data[name] = val
                setattr(self, name, val)
        return self

    def _match_rule(self):
        if len(self._meta.rules) == 0:
            raise TypeError("This Pyot object is static")
        for key, attr in self._meta.rules.items():
            load = {}
            for a in attr:
                try:
                    checkonly = a.startswith('?')
                    if checkonly:
                        getattr(self, a[1:])
                        continue
                    load[a] = getattr(self, a)
                except AttributeError:
                    break
            else:
                self._meta.key = key
                self._meta.load = load
                return self
        raise TypeError("Incomplete values for pipeline token creation")

    def _match_server(self):
        server_type = self._meta.server_type
        if server_type:
            server: str = getattr(self, server_type)
            self._meta.server = server.lower()
        else:
            self._meta.server = None
        return self

    def _place_query(self, kwargs: Dict):
        '''Parse and place request query parameters from dict.'''
        kwargs.pop("self", None)
        kwargs.pop("__class__", None)
        self._meta.query = convert_keys_camel_case(kwargs)

    def _place_body(self, kwargs: Dict):
        '''Parse and place request body parameters from dict.'''
        kwargs.pop("self", None)
        kwargs.pop("__class__", None)
        self._meta.body = convert_keys_camel_case(kwargs)

    def _after_request(self, data: Any, force_copy: bool):
        data = self.filter(data)
        self._meta.raw_data = fast_copy(data) if force_copy else data
        self._meta.data = self.transform(data)
        self.fill()

    async def setup(self):
        """Coroutine. Set up the object to make request, this comes before `validate()`."""

    async def token(self) -> PipelineToken:
        '''Coroutine. Create a pipeline token that identifies this object (its parameters).'''
        await self.setup()
        self._match_rule()
        self._match_server()
        self.validate()
        try:
            return PipelineToken(self._meta.pipeline.model, self._meta.server, self._meta.key, self._meta.load, self._meta.query)
        except AttributeError as e:
            raise ValueError("Token creation failed, please ensure a pipeline is activated or provided") from e

    async def get(self, force_copy: bool = False):
        '''Coroutine. Make a GET request to the pipeline.'''
        data = await self._meta.pipeline.get(await self.token())
        self._after_request(data, force_copy)
        return self

    async def post(self, force_copy: bool = False):
        '''Coroutine. Make a POST request to the pipeline.'''
        data = await self._meta.pipeline.post(await self.token(), self._meta.body)
        self._after_request(data, force_copy)
        return self

    async def put(self, force_copy: bool = False):
        '''Coroutine. Make a PUT request to the pipeline.'''
        data = await self._meta.pipeline.put(await self.token(), self._meta.body)
        self._after_request(data, force_copy)
        return self

    def using(self, pipeline_name: str):
        '''Set the pipeline used for request.'''
        try:
            self._meta.pipeline = pipelines[pipeline_name]
        except KeyError as e:
            raise ValueError(f"Pipeline '{pipeline_name}' does not exist, inactive or dead") from e
        return self

    def validate(self):
        """Validate the object to make request, this comes right before the request."""

    def filter(self, data):
        """Filter out the requested data from the response, this comes right after the request and before 'transform()'."""
        return data

    def transform(self, data) -> Dict:
        """Transform the data into a pyot compatible structure non-destructively, this comes after `filter()`."""
        return data

    def raw(self):
        """Return the raw response of the request, only available for Core objects"""
        return self._meta.raw_data

    @classmethod
    def load(cls, raw: Any):
        o = cls()
        o._meta.raw_data = raw
        o._meta.data = o.transform(raw)
        o.fill()
        return o


class PyotUtilBase: pass



from datetime import datetime
from typing import List, Iterator, TYPE_CHECKING

from dateutil.parser import parse

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from ..riot.account import Account
    from .card import Deck


## PYOT STATIC OBJECTS

class MatchMetaData(PyotStatic):
    data_version: str
    match_id: str
    participant_puuids: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"participant_puuids"}
        renamed = {"participants": "participant_puuids"}

    @property
    def participants(self) -> List["Account"]:
        from pyot.models.riot import Account
        return [Account(puuid=puuid).using(self.metapipeline.name) for puuid in self.participant_puuids]


class MatchPlayerData(PyotStatic):
    puuid: str
    deck_id: str
    deck_code: str
    factions: List[str]
    game_outcome: str
    order_of_play: int
    win: bool

    class Meta(PyotStatic.Meta):
        raws = {"factions"}

    @property
    def account(self) -> "Account":
        from pyot.models.riot import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)

    @property
    def deck(self) -> "Deck":
        from .card import Deck
        return Deck(code=self.deck_code)


class MatchInfoData(PyotStatic):
    mode: str # (Legal values: Constructed, Expeditions, Tutorial)
    type: str # (Legal values: Ranked, Normal, AI, Tutorial, VanillaTrial, Singleton, StandardGauntlet)
    start_time_strftime: str
    version: str
    players: List[MatchPlayerData]
    total_turn_count: int

    class Meta(PyotStatic.Meta):
        renamed = {"game_mode": "mode", "game_type": "type", "game_start_time_utc": "start_time_strftime", "game_version": "version"}

    @property
    def start_time(self) -> datetime:
        return parse(self.start_time_strftime)


## PYOT CORE OBJECTS

class Match(PyotCore):
    id: str
    metadata: MatchMetaData
    info: MatchInfoData

    class Meta(PyotCore.Meta):
        rules = {"match_v1_match": ["id"]}

    def __init__(self, id: str = empty, region: str = models.lor.DEFAULT_REGION):
        self.initialize(locals())

    def transform(self, data):
        for player in data["info"]["players"]:
            if player["game_outcome"] == "win":
                player["win"] = True
            elif player["game_outcome"] == "loss":
                player["win"] = False
            else:
                player["win"] = None
        return data


class MatchHistory(PyotCore):
    ids: List[str]
    puuid: str

    class Meta(PyotCore.Meta):
        raws = {"ids"}
        rules = {"match_v1_matchlist": ["puuid"]}

    def __init__(self, puuid: str = empty, region: str = models.lor.DEFAULT_REGION):
        self.initialize(locals())

    def transform(self, data):
        return {"ids": data}

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.matches[item]

    def __iter__(self) -> Iterator[Match]:
        return iter(self.matches)

    def __len__(self):
        return len(self.ids)

    @property
    def matches(self) -> List[Match]:
        return [Match(id=id_, region=self.region) for id_ in self.ids]

    @property
    def account(self) -> "Account":
        from pyot.models.riot import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)




from typing import Dict
from pyot.pipeline.token import PipelineToken
from .base import BaseEndpoint


class RiotAPIEndpoint(BaseEndpoint):

    base_url = "https://{server}.api.riotgames.com"

    shared = {
        "account_v1_by_puuid": "/riot/account/v1/accounts/by-puuid/{puuid}",
        "account_v1_by_riot_id": "/riot/account/v1/accounts/by-riot-id/{game_name}/{tag_line}",
        "account_v1_active_shard": "/riot/account/v1/active-shards/by-game/{game}/by-puuid/{puuid}",
    }

    all = {
        "lol": {
            "champion_v3_rotation": "/lol/platform/v3/champion-rotations",
            "champion_mastery_v4_by_champion_id": "/lol/champion-mastery/v4/champion-masteries/by-summoner/{summoner_id}/by-champion/{champion_id}",
            "champion_mastery_v4_all_mastery": "/lol/champion-mastery/v4/champion-masteries/by-summoner/{summoner_id}",
            "clash_v1_players": "/lol/clash/v1/players/by-summoner/{summoner_id}",
            "clash_v1_teams": "/lol/clash/v1/teams/{id}",
            "clash_v1_tournaments_by_team_id": "/lol/clash/v1/tournaments/by-team/{team_id}",
            "clash_v1_toutnaments_by_tournament_id": "/lol/clash/v1/tournaments/{id}",
            "clash_v1_tournaments_all": "/lol/clash/v1/tournaments",
            "league_v4_summoner_entries": "/lol/league/v4/entries/by-summoner/{summoner_id}",
            "league_v4_challenger_league": "/lol/league/v4/challengerleagues/by-queue/{queue}",
            "league_v4_grandmaster_league": "/lol/league/v4/grandmasterleagues/by-queue/{queue}",
            "league_v4_master_league": "/lol/league/v4/masterleagues/by-queue/{queue}",
            "league_v4_entries_by_division": "/lol/league/v4/entries/{queue}/{tier}/{division}",
            "league_v4_league_by_league_id": "/lol/league/v4/leagues/{id}",
            "status_v4_platform_data": "/lol/status/v4/platform-data",
            "match_v4_match": "/lol/match/v4/matches/{id}",
            "match_v4_timeline": "/lol/match/v4/timelines/by-match/{id}",
            "match_v4_matchlist": "/lol/match/v4/matchlists/by-account/{account_id}",
            "match_v4_tournament_match": "/lol/match/v4/matches/{id}/by-tournament-code/{tournament_code}",
            "match_v4_tournament_matches": "/lol/match/v4/matches/by-tournament-code/{tournament_code}/ids",
            "match_v5_match": "/lol/match/v5/matches/{id}",
            "match_v5_timeline": "/lol/match/v5/matches/{id}/timeline",
            "match_v5_matches": "/lol/match/v5/matches/by-puuid/{puuid}/ids",
            "spectator_v4_current_game": "/lol/spectator/v4/active-games/by-summoner/{summoner_id}",
            "spectator_v4_featured_games": "/lol/spectator/v4/featured-games",
            "summoner_v4_by_name": "/lol/summoner/v4/summoners/by-name/{name}",
            "summoner_v4_by_id": "/lol/summoner/v4/summoners/{id}",
            "summoner_v4_by_account_id": "/lol/summoner/v4/summoners/by-account/{account_id}",
            "summoner_v4_by_puuid": "/lol/summoner/v4/summoners/by-puuid/{puuid}",
            "third_party_code_v4_code": "/lol/platform/v4/third-party-code/by-summoner/{summoner_id}",
            "tournament_v4_codes": "/lol/tournament/v4/codes",
            "tournament_v4_codes_by_code": "/lol/tournament/v4/codes/{code}",
            "tournament_v4_lobby_events": "/lol/tournament/v4/lobby-events/by-code/{code}",
            "tournament_v4_providers": "/lol/tournament/v4/providers",
            "tournament_v4_tournaments": "/lol/tournament/v4/tournaments",
            "tournament_stub_v4_codes": "/lol/tournament-stub/v4/codes",
            "tournament_stub_v4_lobby_events": "/lol/tournament-stub/v4/lobby-events/by-code/{code}",
            "tournament_stub_v4_providers": "/lol/tournament-stub/v4/providers",
            "tournament_stub_v4_tournaments": "/lol/tournament-stub/v4/tournaments",
        },
        "tft": {
            "league_v1_summoner_entries": "/tft/league/v1/entries/by-summoner/{summoner_id}",
            "league_v1_challenger_league": "/tft/league/v1/challenger",
            "league_v1_grandmaster_league": "/tft/league/v1/grandmaster",
            "league_v1_master_league": "/tft/league/v1/master",
            "league_v1_entries_by_division": "/tft/league/v1/entries/{tier}/{division}",
            "league_v1_league_by_league_id": "/tft/league/v1/leagues/{id}",
            "match_v1_matchlist": "/tft/match/v1/matches/by-puuid/{puuid}/ids",
            "match_v1_match": "/tft/match/v1/matches/{id}",
            "summoner_v1_by_name": "/tft/summoner/v1/summoners/by-name/{name}",
            "summoner_v1_by_id": "/tft/summoner/v1/summoners/{id}",
            "summoner_v1_by_account_id": "/tft/summoner/v1/summoners/by-account/{account_id}",
            "summoner_v1_by_puuid": "/tft/summoner/v1/summoners/by-puuid/{puuid}",
        },
        "lor": {
            "ranked_v1_leaderboards": "/lor/ranked/v1/leaderboards",
            "match_v1_matchlist": "/lor/match/v1/matches/by-puuid/{puuid}/ids",
            "match_v1_match": "/lor/match/v1/matches/{id}",
            "status_v1_platform_data": "/lor/status/v1/platform-data",
        },
        "val": {
            "match_v1_match": "/val/match/v1/matches/{id}",
            "match_v1_matchlist": "/val/match/v1/matchlists/by-puuid/{puuid}",
            "match_v1_recent": "/val/match/v1/recent-matches/by-queue/{queue}",
            "content_v1_contents": "/val/content/v1/contents",
            "ranked_v1_leaderboards": "/val/ranked/v1/leaderboards/by-act/{act_id}",
            "status_v1_platform_data": "/val/status/v1/platform-data",
        }
    }

    def clean(self, token: PipelineToken) -> Dict[str, str]:
        return {
            "server": token.server,
            **token.params,
        }




from typing import List, Mapping, TypeVar, Type, Union
from importlib import import_module


K = TypeVar("K")
V = TypeVar("V")


class ConfDict(dict, Mapping[K, V]):

    def __init__(self, cls: Type[V], throw: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.throw = throw

    def __getattr__(self, k: K) -> V:
        try:
            return self[k]
        except KeyError as e:
            if self.throw:
                raise AttributeError(self.throw.format(k)) from e
            return None

    def __setattr__(self, name: str, value: V) -> None:
        self[name] = value


def reraise_model_inactive(func):

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise RuntimeError("Model partially activated") from e

    return wrapper


def import_confs(path_or_paths: Union[str, List[str]]):
    if isinstance(path_or_paths, str):
        import_module(path_or_paths)
    elif isinstance(path_or_paths, list):
        for path in path_or_paths:
            import_module(path)
    else:
        raise ValueError(path_or_paths)



from typing import List, Iterator

from pyot.conf.model import models
from .base import PyotCore, PyotStatic

## PYOT STATIC OBJECTS

class LeaderboardPlayerData(PyotStatic):
    name: str
    rank: int
    lp: float


# PYOT CORE OBJECTS

class Leaderboard(PyotCore):
    players: List[LeaderboardPlayerData]

    class Meta(PyotCore.Meta):
        rules = {"ranked_v1_leaderboards": []}

    def __init__(self, region: str = models.lor.DEFAULT_REGION):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.players[item]

    def __iter__(self) -> Iterator[LeaderboardPlayerData]:
        return iter(self.players)

    def __len__(self) -> int:
        return len(self.players)



from concurrent.futures import ThreadPoolExecutor
import warnings

from pyot.core.resources import resource_manager, resource_managed_loops
from pyot.core.queue import Queue
from pyot.core.warnings import PyotResourceWarning
from pyot.models import lol
from pyot.utils.sync import async_to_sync

from .engine_core import assert_types, assert_walkable


@async_to_sync
async def test_resource_and_queue():
    async with resource_manager():
        assert len(resource_managed_loops) == 1
        async with Queue() as queue:
            o = await lol.Summoner(name="Morimorph").get()
            h = await o.match_history.get()
            for m in h.matches[:5]:
                await queue.put(m.get())
            ms = await queue.join()
        assert len(ms) == 5
        assert_walkable(o)
        assert_types(o)
    assert len(resource_managed_loops) == 0


@async_to_sync
async def test_resource():
    async with resource_manager():
        o = await lol.Summoner(name="Morimorph").get()
    assert len(resource_managed_loops) == 0
    assert_walkable(o)
    assert_types(o)


def test_resource_multithread():

    @async_to_sync
    async def inner():
        async with resource_manager(), Queue() as queue:
            o = await lol.Summoner(name="Morimorph").get()
            h = await o.match_history.get()
            for m in h.matches[:5]:
                await queue.put(m.get())
            ms = await queue.join()
        assert len(ms) == 5
        for o in ms:
            assert_walkable(o)
            assert_types(o)

    @async_to_sync
    @resource_manager.as_decorator
    async def inner_decorated():
        async with Queue() as queue:
            o = await lol.Summoner(name="Morimorph").get()
            h = await o.match_history.get()
            for m in h.matches[:5]:
                await queue.put(m.get())
            ms = await queue.join()
        assert len(ms) == 5
        for o in ms:
            assert_walkable(o)
            assert_types(o)

    warnings.filterwarnings("error")
    futures = []
    with ThreadPoolExecutor() as executor:
        for _ in range(3):
            futures.append(executor.submit(inner))
            futures.append(executor.submit(inner_decorated))
        for future in futures:
            future.result()
    warnings.filterwarnings("default")


def test_resource_multithread_fail_missing():

    @async_to_sync
    async def inner():
        async with Queue() as queue:
            o = await lol.Summoner(name="Morimorph").get()
            h = await o.match_history.get()
            for m in h.matches[:5]:
                await queue.put(m.get())
            ms = await queue.join()
        assert len(ms) == 5
        for o in ms:
            assert_walkable(o)
            assert_types(o)

    warnings.filterwarnings("error")
    futures = []
    with ThreadPoolExecutor() as executor:
        for _ in range(5):
            futures.append(executor.submit(inner))
        for future in futures:
            try:
                future.result()
            except PyotResourceWarning:
                assert True
    warnings.filterwarnings("default")


def test_resource_multithread_fail_exists():

    async def inner_squared(m: lol.Match):
        async with resource_manager():
            return await m.get()

    @async_to_sync
    async def inner():
        async with resource_manager(), Queue() as queue:
            o = await lol.Summoner(name="Morimorph").get()
            h = await o.match_history.get()
            o = await inner_squared(h.matches[0])
        assert_walkable(o)
        assert_types(o)

    warnings.filterwarnings("error")
    futures = []
    with ThreadPoolExecutor() as executor:
        for _ in range(5):
            futures.append(executor.submit(inner))
        for future in futures:
            try:
                future.result()
                assert False
            except RuntimeError as e:
                assert "already" in str(e)
    warnings.filterwarnings("default")




from pyot.models import val
from pyot.utils.sync import async_to_sync

from .engine_core import assert_types, assert_walkable


@async_to_sync
async def test_content():
    o = await val.Content().query(locale="en-US").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_ranked():
    c = await val.Content().get()
    latest_act = next(act for act in reversed(c.acts) if act.is_active and act.type == 'act')
    act_id = latest_act.id
    o = await val.Leaderboard(act_id=act_id, platform="eu").get()
    assert_walkable(o)
    assert_types(o)
    o = await val.Leaderboard(act_id=act_id, platform="eu").query(size=200, start_index=220).get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_match():
    c = await val.Content().get()
    latest_act = next(act for act in reversed(c.acts) if act.is_active and act.type == 'act')
    act_id = latest_act.id
    l = await val.Leaderboard(act_id=act_id, platform="eu").get()
    o = await val.MatchHistory(puuid=l[0].puuid, platform="eu").get()
    assert_walkable(o)
    assert_types(o)
    for ind in range(2):
        o_ = await o[ind].get()
        assert_walkable(o_)
        assert_types(o_)


@async_to_sync
async def test_status():
    o = await val.Status(platform="eu").get()
    assert_walkable(o)
    assert_types(o)



from pyot.models import tft
from pyot.utils.sync import async_to_sync
from pyot.utils.tft.routing import platform_to_region

from .engine_core import assert_types, assert_walkable


@async_to_sync
async def test_champion():
    o = await tft.Champions(version="pbe", locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.Champion(key=o[0].key, locale="es_mx").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_item():
    o = await tft.Items(version="pbe", locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.Item(id=o[0].id, version="pbe", locale="es_mx").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_league():
    o = await tft.ChallengerLeague(platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.GrandmasterLeague(platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.MasterLeague(platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.League(id=o.id, platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    s = await tft.Summoner(name="Morimorph", platform="na1").get()
    o = await tft.SummonerLeague(summoner_id=s.id, platform="na1").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.DivisionLeague(tier="GOLD", division="I", platform="na1").query(page=1).get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_match():
    s = await tft.Summoner(name="Morimorph", platform="na1").get()
    o = await tft.MatchHistory(puuid=s.puuid, region=platform_to_region(s.platform)).query(count=100).get()
    assert_walkable(o)
    assert_types(o)
    o = await s.match_history.query(count=100).get()
    assert_walkable(o)
    assert_types(o)
    for ind in range(2):
        i = o[ind].id
        o_ = await o[ind].get()
        assert_walkable(o_)
        assert_types(o_)
        o1 = await tft.Match(id=i, region=platform_to_region(s.platform)).get()
        assert_walkable(o1)
        assert_types(o1)


@async_to_sync
async def test_profileicon():
    o = await tft.ProfileIcons(locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await o[230].get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_summoner():
    o = await tft.Summoner(name="Morimorph").get()
    assert_walkable(o)
    assert_types(o)


@async_to_sync
async def test_trait():
    o = await tft.Traits(version="pbe", locale="en_us").get()
    assert_walkable(o)
    assert_types(o)
    o = await tft.Trait(key=o[0].key, version="pbe", locale="es_mx").get()
    assert_walkable(o)
    assert_types(o)



from abc import ABC
import warnings

from pyot.core.warnings import PyotConfWarning

from .utils import ConfDict, reraise_model_inactive


AVAILABLE_MODELS = {"riot", "lol", "tft", "lor", "val"}


class ModelConf(ABC):

    default_platform: str = None
    default_region: str = None
    default_locale: str = None
    default_version: str = None


class Model:

    def __init__(self, default_platform: str, default_region: str, default_locale: str, default_version: str) -> None:
        self.default_platform = default_platform
        self.default_region = default_region
        self.default_locale = default_locale
        self.default_version = default_version

    @property
    @reraise_model_inactive
    def DEFAULT_PLATFORM(self):
        return self.default_platform

    @property
    @reraise_model_inactive
    def DEFAULT_REGION(self):
        return self.default_region

    @property
    @reraise_model_inactive
    def DEFAULT_LOCALE(self):
        return self.default_locale

    @property
    @reraise_model_inactive
    def DEFAULT_VERSION(self):
        return self.default_version


models: ConfDict[str, Model] = ConfDict(Model, "Model '{0}' is inactive or does not exist")


def activate_model(name: str):
    '''Make the settings take effect.'''

    keys = {"default_platform", "default_region", "default_locale", "default_version"}

    if name not in AVAILABLE_MODELS:
        raise ValueError(f"Invalid model, '{name}' was given")

    def wrapper(cls: "ModelConf"):
        for key in keys:
            if hasattr(cls, key):
                continue
            raise ValueError(f"Missing value for '{key}' in {cls} conf")
        if name in models or hasattr(models, name):
            warnings.warn(f"An attempt to activate model '{name}' was ignored due to model already active", PyotConfWarning)
            return cls
        models[name] = Model(cls.default_platform, cls.default_region, cls.default_locale, cls.default_version)
        return cls

    return wrapper



from typing import List, Iterator

from pyot.conf.model import models
from pyot.core.functional import cache_indexes, lazy_property, empty
from pyot.utils.lol.cdragon import abs_url
from .base import PyotCore


class ProfileIcon(PyotCore):
    id: int
    icon_path: str

    class Meta(PyotCore.Meta):
        rules = {"cdragon_profile_icon_full": ["version", "locale", "?id"]}

    def __init__(self, id: int = empty, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(self.id, data, "id")

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)


class ProfileIcons(PyotCore):
    icons: List[ProfileIcon]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_profile_icon_full": ["version", "locale"]}

    def __init__(self, version: str = models.lol.DEFAULT_VERSION, locale: str = models.lol.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.icons[item]

    def __iter__(self) -> Iterator[ProfileIcon]:
        return iter(self.icons)

    def __len__(self):
        return len(self.icons)

    def transform(self, data):
        return {"icons": data}



from typing import Dict, List, Iterator, Union

from lor_deckcodes.encode import encode_deck
from lor_deckcodes.decode import decode_deck

from pyot.conf.model import models
from pyot.core.objects import PyotUtilBase
from pyot.core.functional import cache_indexes, empty
from pyot.utils.lor.cards import batch_to_ccac
from .base import PyotCore, PyotStatic


# PYOT STATIC OBJECTS

class CardAssetData(PyotStatic):
    game_absolute_path: str
    full_absolute_path: str


# PYOT CORE OBJECTS

class Card(PyotCore):
    associated_card_codes: List[str]
    associated_card_refs: List[str]
    assets: List[CardAssetData]
    region: str = None
    region_ref: str
    regions: List[str]
    region_refs: List[str]
    attack: int
    cost: int
    health: int
    description: str
    description_raw: str
    levelup_description: str
    levelup_description_raw: str
    flavor_text: str
    artist_name: str
    name: str
    code: str
    keywords: List[str]
    keyword_refs: List[str]
    spell_speed: str
    spell_speed_ref: str
    rarity: str
    rarity_ref: str
    subtype: str
    subtypes: List[str]
    supertype: str
    type: str
    collectible: bool
    set: int
    faction: str
    number: int
    subcode: str

    class Meta(PyotCore.Meta):
        raws = {"keywords", "keyword_refs", "subtypes", "associated_card_codes", "associated_card_refs", "regions", "region_refs"}
        renamed = {"card_code": "code", "associated_cards": "associated_card_codes"}
        rules = {"ddragon_lor_set_data": ["set", "?code", "version", "locale"]}

    def __init__(self, code: str = empty, version: str = models.lor.DEFAULT_VERSION, locale: str = models.lor.DEFAULT_LOCALE):
        self.initialize(locals())
        if code:
            self.set = int(code[:2])
            self.faction = code[2:4]
            self.number = int(code[4:])

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(self.code, data, "cardCode")

    def transform(self, data):
        data["set"] = int(data["set"][3:])
        data["faction"] = data["cardCode"][2:4]
        data["number"] = int(data["cardCode"][4:7])
        data["subcode"] = data["cardCode"][7:] if len(data["cardCode"]) > 7 else ""
        return data

    def __str__(self):
        return self.code

    @property
    def associated_cards(self) -> List["Card"]:
        return [Card(code=code, version=self.version, locale=self.locale) for code in self.associated_card_codes]


class Cards(PyotCore):
    cards: List[Card]

    class Meta(PyotCore.Meta):
        rules = {"ddragon_lor_set_data": ["set", "version", "locale"]}

    def __init__(self, set: int = empty, version: str = models.lor.DEFAULT_VERSION, locale: str = models.lor.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.cards[item]

    def __iter__(self) -> Iterator[Card]:
        return iter(self.cards)

    def __len__(self):
        return len(self.cards)

    def transform(self, data):
        return {"cards": data}


## PYOT CONTAINERS

class Batch(PyotUtilBase):
    code: str
    count: int
    faction: str
    set: int
    number: int

    def __init__(self, code: str = empty, count: int = 1, raw: str = empty):
        if code:
            self.code = code
            self.count = int(count)
        elif raw:
            card_code_and_count = raw.split(":")
            self.code = card_code_and_count[1]
            self.count = card_code_and_count[0]
        else:
            raise RuntimeError("Batch takes at least 'code' or 'raw' string, prioritizing 'code'")
        self.set = int(self.code[:2])
        self.faction = self.code[2:4]
        self.number = int(self.code[4:])

    def __str__(self):
        return f"{self.count}:{self.code}"

    def add(self, amount: int = 1):
        '''Add a copy to the batch, `amount` may be passed to add more than 1 copy.'''
        if self.count + amount <= 3:
            self.count += amount
        else:
            raise RuntimeError("The batch cannot contain more than 3 copies in the current version")

    def remove(self, amount: int = 1):
        '''Remove a copy from the batch, `amount` may be passed to remove more than 1 copy.'''
        if self.count > amount:
            self.count -= 1
        else:
            raise RuntimeError("The batch needs to have at least 1 copy, delete it instead")

    def dict(self):
        return {
            "code": self.code,
            "count": self.count,
            "faction": self.faction,
            "set": self.set,
            "number": self.number,
        }

    @property
    def card(self) -> Card:
        return Card(code=self.code)


class Deck(PyotUtilBase):
    batches: List[Batch]
    code: str

    def __init__(self, batches: Union[List[str], List[Batch]] = empty, code: str = empty):
        self.batches = []
        if batches:
            for batch in batches:
                self.append(batch)
        if code:
            self.code = code

    def __getitem__(self, item):
        if not isinstance(item, int):
            return getattr(self, item)
        return self.batches[item]

    def __iter__(self):
        return iter(self.batches)

    def __len__(self) -> List[Batch]:
        return len(self.batches)

    def append(self, batch: Union[Batch, str]):
        '''Appends a Batch object or CardCodeAndCount string to the Deck.'''
        if isinstance(batch, Batch):
            self.batches.append(batch)
            return
        card_code_and_count = batch.split(":")
        self.batches.append(Batch(code=card_code_and_count[1], count=card_code_and_count[0]))

    def pop(self, ind: int = -1) -> Batch:
        '''Remove and return a Batch object by index.'''
        return self.batches.pop(ind)

    def pull(self, card_code: str):
        '''Remove and return a Batch object by code.'''
        for ind, batch in enumerate(self.batches):
            if batch.code == card_code:
                return self.batches.pop(ind)

    def encode(self) -> str:
        '''Encode the content in `self.batches`, set the code and return it.'''
        self.code = encode_deck([batch_to_ccac(batch) for batch in self.batches])
        return self.code

    def decode(self):
        '''Decode the string in `self.code`, rebuild the batches and return self.'''
        lor_deck = decode_deck(self.code)
        self.batches = []
        for lor_card in lor_deck:
            self.append(lor_card)
        return self

    def dict(self) -> List[Dict]:
        return [batch.dict() for batch in self.batches]

    @property
    def raw(self) -> List[str]:
        return [str(batch) for batch in self.batches]



from .tournament import Tournament, TournamentProvider, TournamentCodes, TournamentCode, TournamentLobbyEvents, TournamentStub, TournamentStubCodes, TournamentStubLobbyEvents, TournamentStubProvider
from .championmastery import ChampionMasteries, ChampionMastery
from .league import League, ChallengerLeague, GrandmasterLeague, MasterLeague, SummonerLeague, DivisionLeague
from .clash import ClashPlayers, ClashTeam, ClashTournaments, ClashTournament
from .match import Match, Timeline, MatchHistory
from .profileicon import ProfileIcon, ProfileIcons
from .spectator import FeaturedGames, CurrentGame
from .championrotation import ChampionRotation
from .thirdpartycode import ThirdPartyCode
from .merakichampion import MerakiChampion
from .merakiitem import MerakiItem
from .spell import Spell, Spells
from .champion import Champion, Champions
from .summoner import Summoner
from .item import Item, Items
from .rune import Rune, Runes
from .status import Status



import asyncio
import time

from pyot.models import lol
from pyot.utils.copy import fast_copy
from pyot.utils.functools import async_property, async_cached_property, async_generator_property
from pyot.utils.importlib import import_variable
from pyot.utils.itertools import frozen_generator
from pyot.utils.sync import async_to_sync, sync_to_async


def test_copy_fast_copy():
    o = {"1": 2, "3": 4}
    assert fast_copy(o) == o
    o = lol.Summoner(name="Morimorph")
    assert fast_copy(o).name == o.name


def test_importlib_import_variable():
    assert import_variable("pyot.models.lol.Match") is lol.Match
    assert import_variable("pyot.utils.importlib.import_variable") is import_variable


def test_itertools_frozen_generator():
    o_s = [lol.Summoner(name="Morimorph")] * 5
    for o in frozen_generator(o_s):
        assert o.name == "Morimorph"
        o.name = "SomethingElse"
    for o in frozen_generator(o_s):
        assert o.name == "Morimorph"


@async_to_sync
async def test_sync_sync_to_async():

    @sync_to_async
    def sleeper(secs):
        time.sleep(secs)

    start = time.time()
    await asyncio.gather(*[sleeper(i) for i in range(1, 5)])
    assert time.time() - start < 4.5


@async_to_sync
async def test_functools_async_properties():
    
    class A:
        b_count = 0
        @async_property
        async def a(self):
            return 0
        @async_cached_property
        async def b(self):
            self.b_count += 1
            return self.b_count
        @async_generator_property
        async def c(self):
            for i in range(10):
                yield i
        @async_generator_property
        async def d(self):
            for i in range(10):
                yield i, i * 2

    a = A()
    assert await a.a == 0
    assert await a.b == 1
    assert await a.b == 1
    assert await a.b == 1
    c_idx = 0
    async for c_i in a.c:
        assert c_i == c_idx
        c_idx += 1
    d_idx = 0
    async for d_i, d_i_2 in a.d:
        assert d_i == d_idx
        assert d_i_2 == d_idx * 2
        d_idx += 1



from datetime import datetime
from typing import List, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore

if TYPE_CHECKING:
    from .summoner import Summoner
    from .champion import Champion
    from .merakichampion import MerakiChampion


# PYOT CORE OBJECTS

class ChampionMastery(PyotCore):
    champion_id: int
    champion_level: int
    champion_points: int
    last_play_timestamp: int
    champion_points_since_last_level: int
    champion_points_until_next_level: int
    chest_granted: bool
    tokens_earned: int
    summoner_id: str

    class Meta(PyotCore.Meta):
        rules = {"champion_mastery_v4_by_champion_id": ["summoner_id", "champion_id"]}
        renamed = {'last_play_time': 'last_play_timestamp'}

    def __init__(self, summoner_id: str = empty, champion_id: int = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def last_play_time(self) -> datetime:
        return datetime.fromtimestamp(self.last_play_timestamp//1000)

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(id=self.champion_id)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.champion_id)



class ChampionMasteries(PyotCore):
    summoner_id: str
    masteries: List[ChampionMastery]
    total_score: int

    class Meta(PyotCore.Meta):
        rules = {"champion_mastery_v4_all_mastery": ["summoner_id"]}

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.masteries[item]

    def __iter__(self) -> Iterator[ChampionMastery]:
        return iter(self.masteries)

    def __len__(self):
        return len(self.masteries)

    def __init__(self, summoner_id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        new_data = {}
        new_data["totalScore"] = 0
        for a in data:
            new_data["totalScore"] += a["championLevel"]
        new_data["masteries"] = data
        return new_data

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)




import re


def snake_case(attr: str, sep_numbers=False) -> str:
    '''Convert string to python snake_case.'''
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', attr)
    snake_case = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()
    if sep_numbers:
        snake_case = re.sub('([a-z])([0-9])', r'\1_\2', snake_case)
    return snake_case


def camel_case(snake_str: str) -> str:
    '''Convert string to json camel_case.'''
    components = snake_str.split('_')
    if len(components) == 1:
        return components[0]
    return components[0] + ''.join(x.title() for x in components[1:])



import datetime
from typing import Any
import pickle
import copy

from pyot.utils.threading import AsyncLock
from pyot.core.exceptions import NotFound
from pyot.pipeline.token import PipelineToken
from pyot.pipeline.expiration import ExpirationManager
from pyot.utils.logging import LazyLogger

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class Omnistone(Store):

    type = StoreType.CACHE

    def __init__(self, game: str, max_entries: int = 10000, cull_frecuency: int = 2, expirations: Any = None, log_level: int = 0) -> None:
        self._lock = AsyncLock()
        self._cull_lock = [False, datetime.datetime.now()]
        self.game = game
        self.data = dict()
        self.expirations = ExpirationManager(game, expirations)
        self.max_entries = max_entries
        self.cull_frecuency = cull_frecuency
        self.log_level = log_level

    async def set(self, token: PipelineToken, value: Any, **kwargs) -> None:
        timeout = self.expirations.get_timeout(token.method)
        if timeout != 0:
            async with self._lock:
                if timeout != -1:
                    timeout = datetime.timedelta(seconds=timeout)
                if await self._allowed():
                    value = pickle.dumps(value)
                    self.data[token.value] = (value, timeout, datetime.datetime.now(), datetime.datetime.now())
                    LOGGER.log(self.log_level, f"[pyot.stores.omnistone:Omnistone] SET {token.value}")
            if len(self.data) > self.max_entries and await self._allowed():
                async with self._lock:
                    self._cull_lock = [True, datetime.datetime.now()]
                await self.expire()
                if len(self.data) > self.max_entries - self.max_entries/self.cull_frecuency:
                    await self.cull()

    async def get(self, token: PipelineToken, expiring: bool = False, **kwargs) -> Any:
        timeout = self.expirations.get_timeout(token.method)
        if timeout == 0:
            raise NotFound(token.value)
        async with self._lock:
            try:
                item, timeout, entered, _ = self.data[token.value]
            except KeyError as e:
                raise NotFound(token.value) from e
            if not expiring:
                LOGGER.log(self.log_level, f"[pyot.stores.omnistone:Omnistone] GET {token.value}")

            now = datetime.datetime.now()
            if timeout == -1:
                self.data[token.value] = (item, timeout, entered, now)
                item = pickle.loads(item)
                return item
            elif now > entered + timeout:
                try:
                    del self.data[token.value]
                except KeyError: pass
                raise NotFound(token.value)
            else:
                self.data[token.value] = (item, timeout, entered, now)
                item = pickle.loads(item)
                return item

    async def delete(self, token: PipelineToken, **kwargs) -> None:
        try:
            del self.data[token.value]
            LOGGER.log(self.log_level, f"[pyot.stores.omnistone:Omnistone] DELETE {token.value}")
        except KeyError as e:
            raise NotFound(token.value) from e

    async def expire(self, **kwargs):
        for token in self.data:
            await self.get(token, expiring=True)

    async def clear(self, **kwargs):
        async with self._lock:
            self.data = {}
            LOGGER.log(self.log_level, f"[pyot.stores.omnistone:Omnistone] Store cleared successfully")

    async def _allowed(self):
        if not self._cull_lock[0]:
            return True
        elif self._cull_lock[0] and datetime.datetime.now() > self._cull_lock[1] + datetime.timedelta(seconds=30):
            self._cull_lock[0] = False
            return True
        else:
            return False

    async def cull(self):
        async with self._lock:
            data = copy.copy(self.data)
        lru = sorted(data.keys(), key=lambda x: data[x][3])
        for token in lru[:int(len(data)/self.cull_frecuency)]:
            await self.delete(token)
        self._cull_lock[0] = False

    async def contains(self, token: PipelineToken, **kwargs) -> bool:
        try:
            _ = self.data[token.value]
            return True
        except KeyError:
            return False




from typing import Any, Union
from io import FileIO, BytesIO
import json


def loads(content: Union[str, bytes], **kwargs) -> Any:
    '''Same as json.loads with graceful fallback by returning the passed content as is'''
    try:
        return json.loads(content, **kwargs)
    except json.decoder.JSONDecodeError:
        return content

def load(fp: Union[FileIO, BytesIO], **kwargs) -> Any:
    '''Same as json.load with graceful fallback by returning the read content as is'''
    try:
        return json.load(fp, **kwargs)
    except json.decoder.JSONDecodeError:
        return fp.read()



import os

from pyot.conf.pipeline import activate_pipeline, PipelineConf
from pyot.conf.model import activate_model, ModelConf

from .engine_core import inject_guards


inject_guards()


@activate_model("riot")
class RiotModel(ModelConf):
    default_platform = "na1"
    default_region = "americas"
    default_version = "latest"
    default_locale = "en_us"


@activate_model("lol")
class LolModel(ModelConf):
    default_platform = "na1"
    default_region = "americas"
    default_version = "latest"
    default_locale = "en_us"


@activate_pipeline("lol")
class LolPipeline(PipelineConf):
    name = "lol_main"
    default = True
    stores = [
        {
            "backend": "pyot.stores.omnistone.Omnistone",
            "log_level": 30,
            "expirations": {
                "match_v5_timeline": 600,
            }
        },
        {
            "backend": "pyot.stores.rediscache.RedisCache",
            "db": 6,
            "log_level": 30,
            "expirations": {
                "summoner_v4_by_puuid": 600,
            }
        },
        {
            "backend": "pyot.stores.diskcache.DiskCache",
            "directory": "diskcache",
            "log_level": 30,
            "expirations": {
                "match_v5_match": 600,
            }
        },
        {
            "backend": "pyot.stores.mongodb.MongoDB",
            "db": "pyot_test",
            "log_level": 30,
            "expirations": {
                "summoner_v4_by_name": 100,
            }
        },
        {
            "backend": "pyot.stores.merakicdn.MerakiCDN",
            "log_level": 30,
        },
        {
            "backend": "pyot.stores.cdragon.CDragon",
            "log_level": 30,
        },
        {
            "backend": "pyot.stores.riotapi.RiotAPI",
            "log_level": 30,
            "api_key": os.environ["RIOT_API_KEY"],
            "rate_limiter": {
                "db": 7,
                "backend": "pyot.limiters.redis.RedisLimiter"
            }
        }
    ]


@activate_model("tft")
class TftModel(ModelConf):
    default_platform = "na1"
    default_region = "americas"
    default_version = "latest"
    default_locale = "en_us"


@activate_pipeline("tft")
class TftPipeline(PipelineConf):
    name = "tft_main"
    default = True
    stores = [
        {
            "backend": "pyot.stores.omnistone.Omnistone",
            "log_level": 30,
            "expirations": {
                "summoner_v1_by_name": 100,
                "match_v1_match": 600,
            }
        },
        {
            "backend": "pyot.stores.cdragon.CDragon",
            "log_level": 30,
        },
        {
            "backend": "pyot.stores.riotapi.RiotAPI",
            "log_level": 30,
            "api_key": os.environ["TFT_API_KEY"],
        }
    ]


@activate_model("val")
class ValModel(ModelConf):
    default_platform = "na"
    default_region = "americas"


@activate_pipeline("val")
class ValPipeline(PipelineConf):
    name = "val_main"
    default = True
    stores = [
        {
            "backend": "pyot.stores.omnistone.Omnistone",
            "log_level": 30,
            "expirations": {
                "account_v1_by_riot_id": 100,
                "match_v1_match": 600,
            }
        },
        {
            "backend": "pyot.stores.riotapi.RiotAPI",
            "log_level": 30,
            "api_key": os.environ["PYOT_DEV_KEY"],
        }
    ]


@activate_model("lor")
class LorModel(ModelConf):
    default_region = "americas"
    default_version = "latest"
    default_locale = "en_us"


@activate_pipeline("lor")
class LorPipeline(PipelineConf):
    name = "lor_main"
    default = True
    stores = [
        {
            "backend": "pyot.stores.omnistone.Omnistone",
            "log_level": 30,
            "expirations": {
                "account_v1_by_riot_id": 100,
                "match_v1_match": 600,
            }
        },
        {
            "backend": "pyot.stores.ddragon.DDragon",
            "log_level": 30,
        },
        {
            "backend": "pyot.stores.riotapi.RiotAPI",
            "log_level": 30,
            "api_key": os.environ["LOR_API_KEY"],
        }
    ]



from typing import List, Iterator, TYPE_CHECKING
from datetime import datetime, timedelta

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from .champion import Champion
    from .merakichampion import MerakiChampion
    from .rune import Rune
    from .spell import Spell
    from .summoner import Summoner
    from .profileicon import ProfileIcon


# PYOT STATIC OBJECTS

class CurrentGameBansData(PyotStatic):
    pick_turn: int
    champion_id: int
    team_id: int

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(id=self.champion_id)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.champion_id)


class CurrentGameParticipantCustomizationData(PyotStatic):
    category: str
    content: str


class CurrentGameParticipantData(PyotStatic):
    team_id: int
    champion_id: int
    profile_icon_id: int
    is_bot: bool
    summoner_name: str
    summoner_id: str
    spell_ids: List[int]
    rune_ids: List[int]
    rune_main_style: int
    rune_sub_style: int
    game_customization_objects: List[CurrentGameParticipantCustomizationData]
    position: str

    class Meta(PyotStatic.Meta):
        renamed = {"bot": "is_bot", "perk_ids": "rune_ids", "perk_style": "rune_main_style", "perk_sub_style": "rune_sub_style"}
        raws = {"spell_ids", "rune_ids"}

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, name=self.summoner_name, platform=self.platform)

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(id=self.champion_id)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.champion_id)

    @property
    def profile_icon(self) -> "ProfileIcon":
        from .profileicon import ProfileIcon
        return ProfileIcon(id=self.profile_icon_id)

    @property
    def runes(self) -> List["Rune"]:
        from .rune import Rune
        mutable = []
        for i in self.rune_ids:
            mutable.append(Rune(id=i))
        return mutable

    @property
    def spells(self) -> List["Spell"]:
        from .spell import Spell
        mutable = []
        for i in self.spell_ids:
            mutable.append(Spell(id=i))
        return mutable


class FeaturedGameParticipantData(PyotStatic):
    team_id: int
    champion_id: int
    profile_icon_id: int
    is_bot: bool
    summoner_name: str
    spell_ids: List[int]
    position: str

    class Meta(PyotStatic.Meta):
        renamed = {"bot": "is_bot"}
        raws = {"spell_ids"}

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(name=self.summoner_name, platform=self.platform)

    @property
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(id=self.champion_id)

    @property
    def meraki_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.champion_id)

    @property
    def profile_icon(self) -> "ProfileIcon":
        from .profileicon import ProfileIcon
        return ProfileIcon(id=self.profile_icon_id)

    @property
    def spells(self) -> List["Spell"]:
        from .spell import Spell
        mutable = []
        for i in self.spell_ids:
            mutable.append(Spell(id=i))
        return mutable


class CurrentGameTeamData(PyotStatic):
    id: int
    bans: List[CurrentGameBansData]
    participants: List[CurrentGameParticipantData]


class FeaturedGameTeamData(PyotStatic):
    id: int
    bans: List[CurrentGameBansData]
    participants: List[FeaturedGameParticipantData]


class FeaturedGameData(PyotStatic):
    id: int
    type: str
    mode: str
    start_time_millis: int
    length_secs: int #not milliseconds
    map_id: int
    platform: str
    queue_id: int
    observers_key: str
    teams: List[FeaturedGameTeamData]

    class Meta(PyotStatic.Meta):
        renamed = {"game_id": "id", "game_type": "type", "game_start_time": "start_time_millis", "game_mode": "mode",
            "game_length": "length_secs", "platform_id": "platform", "game_queue_config_id": "queue_id"}

    @property
    def start_time(self) -> datetime:
        return datetime.fromtimestamp(self.start_time_millis//1000)

    @property
    def length(self) -> timedelta:
        return timedelta(seconds=self.length_secs)

    @property
    def banned_champions(self) -> List[CurrentGameBansData]:
        bans = []
        for team in self.teams:
            bans += team.bans
        return bans

    @property
    def participants(self) -> List[FeaturedGameParticipantData]:
        participants = []
        for team in self.teams:
            participants += team.participants
        return participants

    @property
    def blue_team(self) -> FeaturedGameTeamData:
        return self.teams[0]

    @property
    def red_team(self) -> FeaturedGameTeamData:
        return self.teams[1]


# PYOT CORE OBJECTS

class CurrentGame(FeaturedGameData, PyotCore):
    summoner_id: str
    teams: List[CurrentGameTeamData]

    class Meta(FeaturedGameData.Meta, PyotCore.Meta):
        rules = {"spectator_v4_current_game": ["summoner_id"]}

    def __init__(self, summoner_id: str = empty, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        data = data.copy()
        data["teams"] = [{"id": 100, "bans": [], "participants": []}, {"id": 200, "bans": [], "participants": []}]
        data["observersKey"] = data.pop("observers", None)["encryptionKey"]
        for ban in data.pop("bannedChampions", []):
            if ban["teamId"] == 100:
                data["teams"][0]["bans"].append(ban)
            else:
                data["teams"][1]["bans"].append(ban)
        for p in data.pop("participants", []).copy():
            p = p.copy()
            p["spellIds"] = [p.pop("spell1Id", None), p.pop("spell2Id", None)]
            p.update(p.pop("perks", {}))
            if p["teamId"] == 100:
                data["teams"][0]["participants"].append(p)
            else:
                data["teams"][1]["participants"].append(p)
        return data

    @property
    def banned_champions(self) -> List[CurrentGameBansData]:
        bans = []
        for team in self.teams:
            bans += team.bans
        return bans

    @property
    def participants(self) -> List[CurrentGameParticipantData]:
        participants = []
        for team in self.teams:
            participants += team.participants
        return participants

    @property
    def blue_team(self) -> CurrentGameTeamData:
        return self.teams[0]

    @property
    def red_team(self) -> CurrentGameTeamData:
        return self.teams[1]

    @property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self.summoner_id, platform=self.platform)


class FeaturedGames(PyotCore):
    games: List[FeaturedGameData]
    refresh_interval_secs: int

    class Meta(PyotCore.Meta):
        rules = {"spectator_v4_featured_games": []}
        renamed = {"client_refresh_interval": "refresh_interval_secs", "game_list": "games"}

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.games[item]

    def __iter__(self) -> Iterator[FeaturedGameData]:
        return iter(self.games)

    def __len__(self):
        return len(self.games)

    def __init__(self, platform: str = models.lol.DEFAULT_PLATFORM):
        self.initialize(locals())

    def transform(self, data):
        data = data.copy()
        data["gameList"] = data["gameList"].copy()
        for ind, game in enumerate(data["gameList"]):
            data["gameList"][ind] = CurrentGame.transform(self, game)
        return data

    @property
    def refresh_interval(self) -> timedelta:
        return timedelta(seconds=self.refresh_interval_secs)



from typing import List, Any, Iterator
from aiohttp import TCPConnector

from pyot.core.exceptions import NotFound, NotFindable, PyotException
from pyot.core.resources import ResourceTemplate
from pyot.stores.base import Store
from pyot.utils.aiohttp import SafeClientSession

from .token import PipelineToken


class Pipeline:
    '''Pipeline containing a sequence of stores for request execution.'''

    def __init__(self, model: str, name: str, stores: List[Store]):
        self.model = model
        self.name = name
        self.stores = stores
        self.sessions = ResourceTemplate(
            acquire_func=lambda: SafeClientSession(connector=TCPConnector(ssl=False)),
            release_func=lambda session: session.close()
        )

    def __iter__(self) -> Iterator[Store]:
        return iter(self.stores)

    def __getitem__(self, item):
        return self.stores[item]

    async def get(self, token: PipelineToken):
        '''Get an object from the stores.'''
        session = await self.sessions.acquire()
        found_in = None
        last_exc = NotFindable
        for store in self.stores:
            try:
                response = await store.get(token, session=session)
                found_in = store
            except (NotImplementedError, NotFindable):
                continue
            except PyotException as e:
                last_exc = e
                continue
            break
        else:
            raise last_exc
        await self.set(token, response, found_in)
        return response

    async def set(self, token: PipelineToken, value: Any, stop=None):
        '''Set an object to stores of type Cache.'''
        session = await self.sessions.acquire()
        for store in self.stores:
            if store is stop: break
            try:
                await store.set(token, value, session=session)
            except NotImplementedError:
                continue

    async def post(self, token: PipelineToken, body: Any):
        '''Post an object to stores of type Service.'''
        session = await self.sessions.acquire()
        last_exc = NotFindable
        for store in self.stores:
            try:
                response = await store.post(token, body, session=session)
            except (NotImplementedError, NotFindable):
                continue
            except PyotException as e:
                last_exc = e
                continue
            break
        else:
            raise last_exc
        return response

    async def put(self, token: PipelineToken, body: Any):
        '''Put an object to stores of type Service.'''
        session = await self.sessions.acquire()
        last_exc = NotFindable
        for store in self.stores:
            try:
                response = await store.put(token, body, session=session)
            except (NotImplementedError, NotFindable):
                continue
            except PyotException as e:
                last_exc = e
                continue
            break
        else:
            raise last_exc
        return response

    async def clear(self):
        '''Clear stores of type Cache.'''
        session = await self.sessions.acquire()
        for store in self.stores:
            try:
                await store.clear(session=session)
            except NotImplementedError:
                continue

    async def expire(self):
        '''Expire stores of type Cache that cannot dynamically expire.'''
        session = await self.sessions.acquire()
        for store in self.stores:
            try:
                await store.expire(session=session)
            except NotImplementedError:
                continue

    async def delete(self, token: PipelineToken):
        '''Delete an object from the stores.'''
        session = await self.sessions.acquire()
        for store in self.stores:
            try:
                await store.delete(token, session=session)
            except (NotImplementedError, NotFound):
                continue

    async def contains(self, token: PipelineToken) -> bool:
        '''Check if an object exist in stores of type Cache.'''
        session = await self.sessions.acquire()
        contains = False
        for store in self.stores:
            try:
                contains = await store.contains(token, session=session)
            except NotImplementedError:
                continue
            if contains: break
        return contains



from logging import Logger


class LazyLogger(Logger):
    '''Lazy logger which its `log()` method will do nothing if level equals 0'''

    def __init__(self, name):
        super().__init__(name)

    def log(self, level, msg):
        '''Same as logging.Logger.log, with a new level (0) to skip logging.'''
        if level == 0:
            return
        super().log(level, msg)




from .base import BaseEndpoint


class MerakiCDNEndpoint(BaseEndpoint):

    base_url = "https://cdn.merakianalytics.com/riot"

    all = {
        "lol": {
            "meraki_champion_by_key": "/lol/resources/latest/en-US/champions/{key}.json",
            "meraki_item_by_id": "/lol/resources/latest/en-US/items/{id}.json"
        }
    }



from typing import get_type_hints
import inspect

from typeguard import check_type

from pyot.core.functional import is_laziable, lazy_property
from pyot.core.objects import PyotCoreBase, PyotLazy, PyotStaticBase

from .engine_exceptions import UntypedAttribute
from .engine_utils import get_module_locals, get_properties, get_forward_refs


def assert_walkable(o: PyotStaticBase):
    o.dict(recursive=True, lazy_props=True)


def assert_types(o: PyotStaticBase):
    if isinstance(o, (list, tuple, set)):
        for o_item in o:
            assert_types(o_item)
        return
    if isinstance(o, dict):
        for o_item in o.values():
            assert_types(o_item)
        return
    forward_refs = {"riot": get_forward_refs("riot")}
    if isinstance(o, PyotStaticBase):
        model = o._meta.pipeline.model
        if model not in forward_refs:
            forward_refs[model] = forward_refs["riot"].copy()
            forward_refs[model].update(get_forward_refs(model))
        for key, typ in get_type_hints(o.__class__).items():
            if key[0] == "_" and key[-1] != "_":
                continue
            attr = getattr(o, key, None)
            if attr is None:
                continue
            check_type(f"{o.__class__.__name__}.{key}", attr, typ)
            assert_types(attr)
        for key, func in get_properties(o):
            attr = getattr(o, key, None)
            if attr is None:
                continue
            return_anno = None
            if isinstance(func, lazy_property):
                return_anno = inspect.signature(func.real_func).return_annotation
            elif isinstance(func, property):
                return_anno = inspect.signature(func.fget).return_annotation
            check_type(f"{o.__class__.__name__}.{key}", attr, return_anno, globals=forward_refs[model], locals=get_module_locals(o.__class__))


def inject_guards():
    
    def PyotStaticBasefill(self: PyotStaticBase):
        mapping = self._meta.nomcltrs

        for key, val in self._meta.data.items():
            try:
                attr = mapping[key]
            except KeyError:
                attr = self.qualkey(key)
                mapping[key] = attr

            if attr not in self._meta.types:
                raise UntypedAttribute(self.__class__, attr, val)

            if is_laziable(val):
                if attr in self._meta.raws:
                    setattr(self, attr, val)
                    continue
                try:
                    attr_type = self._meta.types[attr]
                    setattr(self, '_lazy__' + attr, PyotLazy(attr_type[1], attr_type[0], val, self._meta.root))
                except KeyError:
                    setattr(self, attr, val)
            else:
                setattr(self, attr, val)

        return self
    
    PyotStaticBase.fill = PyotStaticBasefill

    def PyotLazy__call__(self: PyotLazy):
        try:
            if issubclass(self.clas, PyotCoreBase):
                if self.container is None:
                    return self.load_core(self.obj)
                if issubclass(self.container, list):
                    return [self.load_core(obj) for obj in self.obj]
                if issubclass(self.container, dict):
                    return {okey: self.load_core(obj) for okey, obj in self.obj.items()}
                raise TypeError(f"Invalid container type '{self.container.__name__}'")
            if self.container is None:
                return self.load_static(self.obj)
            if issubclass(self.container, list):
                return [self.load_static(obj) for obj in self.obj]
            if issubclass(self.container, dict):
                return {okey: self.load_static(obj) for okey, obj in self.obj.items()}
            raise TypeError(f"Invalid container type '{self.container.__name__}'")
        except UntypedAttribute as e:
            raise
        except Exception as e:
            raise RuntimeError(f"Failed to lazy load '{self.clas.__name__}' object due to: ({type(e)}) {e}") from e

    PyotLazy.__call__ = PyotLazy__call__



from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore


class Account(PyotCore):
    puuid: str
    game_name: str
    tag_line: str

    class Meta(PyotCore.Meta):
        rules = {
            "account_v1_by_puuid": ["puuid"],
            "account_v1_by_riot_id": ["game_name", "tag_line"],
        }

    def __init__(self, puuid: str = empty, game_name: str = empty, tag_line: str = empty, region: str = models.riot.DEFAULT_REGION):
        self.initialize(locals())

    def active_shard(self, game: str):
        return ActiveShard(puuid=self.puuid, game=game).using(self.metapipeline.name)


class ActiveShard(PyotCore):
    puuid: str
    game: str
    active_shard: str

    class Meta(PyotCore.Meta):
        rules = {"account_v1_active_shard": ["puuid", "game"]}

    def __init__(self, puuid: str = empty, game: str = empty, region: str = models.riot.DEFAULT_REGION):
        self.initialize(locals())

    @property
    def platform(self):
        try:
            return self.active_shard
        except AttributeError:
            return super().platform

    @property
    def account(self):
        return Account(puuid=self.puuid)




from typing import Any, Dict, Generic, List, Set, TypeVar, Callable, Awaitable, Union
import asyncio
import functools
import inspect
import warnings

from pyot.utils.threading import AsyncLock

from .warnings import PyotResourceWarning


R = TypeVar("R")
F = TypeVar("F", bound=Callable)


resource_managed_loops: Set[asyncio.AbstractEventLoop] = set()
resource_templates: List["ResourceTemplate"] = []


class ResourceManager:
    '''
    Ensures acquisition and releasing of resources used by Pyot.
    - `exist_ok`: > If another resource manager is currently active in the event loop, skip this context,
    avoid using this flag unless unavoidable.
    '''

    def __init__(self, exist_ok: bool = False) -> None:
        self._skip_context = False
        self.loop = asyncio.get_event_loop()
        self.exist_ok = exist_ok

    async def __aenter__(self) -> "ResourceManager":
        current_loop = asyncio.get_event_loop()
        if self.loop != current_loop:
            raise RuntimeError(f"Resource manager bound to event loop {id(self.loop)} cannot acquire using event loop {id(current_loop)}")
        if self.loop in resource_managed_loops:
            if not self.exist_ok:
                raise RuntimeError(f"Event loop {id(self.loop)} already has an active resource manager")
            self._skip_context = True
        else:
            resource_managed_loops.add(self.loop)
            for template in resource_templates:
                await template.acquire(self.loop)
        return self

    async def __aexit__(self, *args) -> None:
        if self._skip_context:
            self._skip_context = False
        else:
            resource_managed_loops.remove(self.loop)
            for template in resource_templates:
                await template.release(self.loop)

    @classmethod
    def as_decorator(cls, func: F) -> F:
        if not inspect.iscoroutinefunction(func):
            raise TypeError(f"{func} is not a coroutine function")
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            async with cls():
                return await func(*args, **kwargs)
        return wrapper

    async def acquire(self):
        return await self.__aenter__()

    async def release(self):
        return await self.__aexit__()


resource_manager = ResourceManager


class ResourceTemplate(Generic[R]):
    '''
    Template for acquiring resources bound to event loops.

    - `acquire_func`: Function for acquiring the resource, the return value will be awaited if it is a coroutine.
    - `release_func`: Function for releasing the resource, the return value will be awaited if it is a coroutine. Optional.

    The submitted functions **must not implement locks**, it may cause deadlocks
    because the acquisition and releasing process are also behind a lock.
    '''

    _acquisitions: Dict[asyncio.AbstractEventLoop, R]

    def __init__(self, acquire_func: Callable[[], Union[R, Awaitable[R]]], release_func: Callable[[R], Any] = None) -> None:
        self._acquisitions = {}
        self._lock = AsyncLock()
        self.acquire_func = acquire_func
        self.release_func = release_func
        resource_templates.append(self)

    async def acquire(self, loop: asyncio.AbstractEventLoop = None) -> R:
        if loop is None:
            loop = asyncio.get_event_loop()
        if loop in self._acquisitions:
            return self._acquisitions[loop]
        async with self._lock:
            if loop in self._acquisitions:
                return self._acquisitions[loop]
            instance = self.acquire_func()
            if inspect.isawaitable(instance):
                self._acquisitions[loop] = await instance
            else:
                self._acquisitions[loop] = instance
            if len(self._acquisitions) > 1 and loop not in resource_managed_loops:
                warnings.warn(
                    f"Multithreading detected but event loop {id(loop)} is not using a resource manager, "
                    "acquired resources by this event loop will not be able to release gracefully.",
                    PyotResourceWarning
                )
            if len(self._acquisitions) % 32 == 0:
                await self.purge()
            return self._acquisitions[loop]

    async def release(self, loop: asyncio.AbstractEventLoop = None):
        if loop is None:
            loop = asyncio.get_event_loop()
        async with self._lock:
            try:
                acquisition = self._acquisitions.pop(loop)
                if self.release_func is not None:
                    called = self.release_func(acquisition)
                    if inspect.isawaitable(called):
                        await called
            except KeyError:
                raise ValueError(f"Acquisition not found for event loop {id(loop)}")

    async def purge(self) -> None:
        '''Purge acquired resources for all closed loops. Ungraceful release.'''
        closed_loops = []
        for loop in self._acquisitions:
            if loop.is_closed():
                closed_loops.append(loop)
        await asyncio.gather(*(self.release(closed_loop) for closed_loop in closed_loops))



import functools
from typing import Awaitable, Callable, TypeVar
from functools import wraps
import inspect
import asyncio


R = TypeVar("R")


def async_to_sync(func: Callable[..., Awaitable[R]]) -> Callable[..., R]:
    '''Wraps `asyncio.run` on an async function converting it into a blocking function. Can be used as decorator @async_to_sync'''
    if not inspect.iscoroutinefunction(func):
        raise TypeError(f"{func} is not a coroutine function")
    @wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))
    return wrapper


def sync_to_async(func: Callable[..., R]) -> Callable[..., Awaitable[R]]:
    '''Wraps `asyncio.get_event_loop().run_in_executor` on a blocking function converting it into a Future. Can be used as decorator @sync_to_async'''
    @wraps(func)
    async def wrapper(*args, **kwargs):
        new_func = functools.partial(func, *args, **kwargs)
        return await asyncio.get_event_loop().run_in_executor(None, new_func)
    return wrapper



from typing import List, Iterator, Dict, Union

from pyot.conf.model import models
from pyot.core.functional import cache_indexes, lazy_property, empty
from pyot.utils.tft.cdragon import sanitize_item_description, abs_url
from .base import PyotCore


# PYOT CORE OBJECT

class Item(PyotCore):
    description: str
    effects: Dict[str, Union[float, str]]
    from_ids: List[int]
    icon_path: str
    id: int
    name: str
    unique: bool

    class Meta(PyotCore.Meta):
        raws = {"from_ids", "effects"}
        rules = {"cdragon_tft_full": ["?id", "version", "locale"]}
        renamed = {"from":"from_ids", "desc": "description", "icon": "icon_path"}

    def __init__(self, id: int = empty, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())

    @cache_indexes
    def filter(self, indexer, data):
        return indexer.get(self.id, data["items"], "id")

    @lazy_property
    def icon_abspath(self) -> str:
        return abs_url(self.icon_path, self.metaroot.version)

    @lazy_property
    def cleaned_description(self) -> str:
        return sanitize_item_description(self.description, self.effects)

    @property
    def from_items(self) -> List["Item"]:
        return [Item(id=i, version=self.version, locale=self.locale) for i in self.from_ids]


class Items(PyotCore):
    items: List[Item]

    class Meta(PyotCore.Meta):
        rules = {"cdragon_tft_full": ["version", "locale"]}

    def __init__(self, version: str = models.tft.DEFAULT_VERSION, locale: str = models.tft.DEFAULT_LOCALE):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.items[item]

    def __iter__(self) -> Iterator[Item]:
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def filter(self, data):
        return data["items"]

    def transform(self, data):
        return {"items": data}



from abc import ABC, abstractmethod
import os
from typing import Callable, Iterable, TypeVar, get_type_hints
from pathlib import Path
import shutil
import pkgutil
import inspect

from pyot.core.objects import PyotCoreBase, PyotRoutingBase, PyotStaticBase, PyotUtilBase, lazy_property
from pyot.utils.importlib import import_variable, import_module
from pyot.utils.nullsafe import nullsafe as _

from .serializers import PyotDocTypeSerializer
from .utils import get_method_properties, get_method_property_names, newline_join


class DocEngine(ABC):

    BASE_PATH = Path.cwd() / 'docs'

    def __init__(self, path: Path = None) -> None:
        if path:
            self.BASE_PATH = path

    @abstractmethod
    def prepare(self):
        pass

    @abstractmethod
    def build(self):
        pass

    def run(self):
        self.prepare()
        self.build()


class ModelsDocEngine(DocEngine):

    BASE_PATH = Path.cwd() / 'docs' / 'models'
    MODELS = ['riot', 'lol', 'tft', 'lor', 'val']

    def prepare(self):
        '''Activate all models and remove existing docs'''
        from pyot import models
        from pyot.conf.model import activate_model, ModelConf

        for module in pkgutil.iter_modules(models.__path__):
            activate_model(module.name)(type(module.name.upper() + "Model", (ModelConf,), {
                "default_region": f"models.{module.name}.DEFAULT_REGION",
                "default_locale": f"models.{module.name}.DEFAULT_LOCALE",
                "default_platform": f"models.{module.name}.DEFAULT_PLATFORM",
                "default_version": f"models.{module.name}.DEFAULT_VERSION",
            }))
        shutil.rmtree(self.BASE_PATH, ignore_errors=True)
        self.BASE_PATH.mkdir(parents=True, exist_ok=True)
        with open(self.BASE_PATH / 'README.md', 'w+') as f:
            f.writelines(["# Models"])

    def build(self):
        for model in self.MODELS:
            model_detail, model_classes = self.get_model_classes(model)
            (self.BASE_PATH / model).mkdir(parents=True, exist_ok=True)
            with open(self.BASE_PATH / model / 'README.md', 'w+') as f:
                lines = [
                    f"# {model_detail['repr']} \n\n",
                ]
                if model_detail['regions']:
                    lines.extend([
                        '## Routing Regions \n\n',
                        newline_join(['* `' + i + '`' for i in sorted(model_detail['regions'])]) + '\n\n',
                    ])
                if model_detail['platforms']:
                    lines.extend([
                        '## Routing Platforms \n\n',
                        newline_join(['* `' + i + '`' for i in sorted(model_detail['platforms'])]) + '\n\n',
                    ])
                f.writelines(lines)
            for submodule, classcontent in model_classes.items():
                with open(self.BASE_PATH / model / (submodule + '.md'), 'w+') as f:
                    lines = [
                        f'# {submodule.capitalize()} \n\n',
                        f'Module: `pyot.models.{model}.{submodule}` \n\n',
                    ]
                    for classtype, classes in classcontent.items():
                        for clas in classes:
                            clas_detail = self.get_model_class_details(model, submodule, clas)
                            lines.append(f'### _class_ `{clas}`' + '\n')
                            lines.append(f'\nType: `Pyot{classtype.capitalize()}` \n')
                            if clas_detail['extends']:
                                lines.append(f'\nExtends: \n')
                                lines.extend([f'* `{ext}` \n' for ext in clas_detail['extends']])
                            if classtype in ('core', 'util'):
                                lines.append(f'\nDefinitions: \n')
                                for def_name, definition in clas_detail['definitions'].items():
                                    lines.extend([
                                        f"* `{def_name}` -> `{definition['returns']}` \n",
                                        *[f"  * `{arg_name}`: `{arg_type.replace('= None', '= empty')}` \n" for arg_name, arg_type in definition['args'].items()]
                                    ])
                            if clas_detail['endpoints']:
                                lines.append(f'\nEndpoints: \n')
                                for end_name, end_args in clas_detail['endpoints'].items():
                                    lines.append(f"* `{end_name}`: `{end_args}` \n")
                            if clas_detail['query_params']:
                                lines.append(f'\nQuery Params: \n')
                                for q_name, q_type in clas_detail['query_params'].items():
                                    lines.append(f"* `{q_name}`: `{q_type.replace('= None', '= empty')}` \n")
                            if clas_detail['body_params']:
                                lines.append(f'\nBody Params: \n')
                                for q_name, q_type in clas_detail['body_params'].items():
                                    lines.append(f"* `{q_name}`: `{q_type.replace('= None', '= empty')}` \n")
                            if clas_detail['methods']:
                                lines.append(f'\nMethods: \n')
                                for att_name, att_content in clas_detail['methods'].items():
                                    lines.append(f"* _{att_content['type']}_ `{att_name}` -> `{att_content['returns']}` \n")
                                    if att_content['args']:
                                        lines.extend([f"  * `{arg_name}`: `{arg_type}` \n" for arg_name, arg_type in att_content['args'].items()])
                                    if att_content['docs']:
                                        doc_str = '\n  > '.join(inspect.cleandoc(att_content['docs']).split('\n'))
                                        lines.append(f"  > {doc_str} \n")
                            if clas_detail['attributes']:
                                lines.append(f'\nAttributes: \n')
                                for att_name, att_content in clas_detail['attributes'].items():
                                    lines.append(f"* `{att_name}` -> `{att_content['returns']}` \n")
                            if clas_detail['properties']:
                                lines.append(f'\nProperties: \n')
                                for att_name, att_content in clas_detail['properties'].items():
                                    lines.append(f"* _{att_content['type']}_ `{att_name}` -> `{att_content['returns']}` \n")
                            lines.append('\n\n')
                            # lines.append(str(clas_detail) + '\n\n')
                    f.writelines(lines)

    def get_model_classes(self, name: str):
        model = import_module(f"pyot.models.{name}")
        modules = pkgutil.iter_modules(model.__path__)
        module_classes = {}
        module_detail = {}

        for module in modules:
            if module.name == "base":
                submodule = import_module(f"pyot.models.{name}.{module.name}")
                module_detail['regions'] = list(_(submodule.PyotRouting)._regions or [])
                module_detail['platforms'] = list(_(submodule.PyotRouting)._platforms or [])
                module_detail['repr'] = submodule.MODULE_REPR
                continue
            submodule = import_module(f"pyot.models.{name}.{module.name}")
            module_classes[module.name] = {"core": [], "static": [], "utils": []}
            for key, obj in inspect.getmembers(submodule):
                obj: object
                if inspect.isclass(obj) and key[0].upper() + key[1:] == key and not key.startswith("Pyot"):
                    if issubclass(obj, PyotCoreBase):
                        module_classes[module.name]["core"].append(key)
                    elif issubclass(obj, PyotStaticBase):
                        module_classes[module.name]["static"].append(key)
                    elif issubclass(obj, PyotUtilBase):
                        module_classes[module.name]["utils"].append(key)
        return module_detail, module_classes


    def get_model_class_details(self, name: str, submodule: str, clasname: str):
        clas: PyotStaticBase = import_variable(f"pyot.models.{name}.{submodule}.{clasname}")
        o = {}
        for key, typ in get_type_hints(clas).items():
            if key[0] == "_" and key[-1] != "_":
                continue
            o[key] = {
                "args": None,
                "returns": PyotDocTypeSerializer(typ, []).data,
                "type": "attribute",
                "docs": None,
            }
        query_params = {}
        body_params = {}
        if issubclass(clas, PyotCoreBase):
            if hasattr(clas, "query"):
                for paramname, typ in inspect.signature(clas.query).parameters.items():
                    if paramname in ("self"):
                        continue
                    typ_anno = typ.annotation
                    if typ.annotation == inspect._empty:
                        if typ.default != inspect._empty and typ.default is not None:
                            typ_anno = typ.default.__class__
                        else:
                            typ_anno = None
                    query_params[paramname] = PyotDocTypeSerializer(typ_anno, []).data
                    if typ.default != inspect._empty:
                        query_params[paramname] += " = " + str(typ.default)
            if hasattr(clas, "body"):
                for paramname, typ in inspect.signature(clas.body).parameters.items():
                    if paramname in ("self"):
                        continue
                    typ_anno = typ.annotation
                    if typ.annotation == inspect._empty:
                        if typ.default != inspect._empty and typ.default is not None:
                            typ_anno = typ.default.__class__
                        else:
                            typ_anno = None
                    body_params[paramname] = PyotDocTypeSerializer(typ_anno, []).data
                    if typ.default != inspect._empty:
                        body_params[paramname] += " = " + str(typ.default)
        base_class_method_props = get_method_property_names(PyotCoreBase) + get_method_property_names(PyotRoutingBase) + ["query", "body"]
        for key, func in get_method_properties(clas):
            if (key not in base_class_method_props or key in [_(clas).Meta.server_type, "__init__", "__iter__"]) and key not in o:
                member_type = "method"
                args = None
                docs_string = None
                if key.startswith("__") and key.endswith("__"):
                    member_type = "definition"
                    if key == "__init__":
                        docs_string = clas.__doc__
                elif key.startswith("_"):
                    continue
                if isinstance(func, lazy_property):
                    member_type = "lazy_property"
                    return_type = inspect.signature(func.real_func).return_annotation
                    docs_string = func.real_func.__doc__
                elif isinstance(func, property):
                    member_type = "property"
                    return_type = inspect.signature(func.fget).return_annotation
                    docs_string = func.fget.__doc__
                else:
                    return_type = inspect.signature(func).return_annotation
                    docs_string = func.__doc__
                    args = {}
                    for paramname, typ in inspect.signature(func).parameters.items():
                        if paramname == "self":
                            continue
                        typ_anno = typ.annotation
                        if typ.annotation == inspect._empty:
                            if typ.default != inspect._empty and typ.default is not None:
                                typ_anno = typ.default.__class__
                            else:
                                typ_anno = None
                        args[paramname] = PyotDocTypeSerializer(typ_anno, []).data
                        if typ.default != inspect._empty:
                            args[paramname] += " = " + str(typ.default)
                if return_type is inspect._empty:
                    return_type = None
                if key == "__str__":
                    return_type = "str"
                if key == "__len__":
                    return_type = "int"
                if inspect.iscoroutinefunction(func):
                    member_type = "async" + member_type
                o[key] = {
                    "args": args,
                    "returns": PyotDocTypeSerializer(return_type, []).data,
                    "type": member_type,
                    "docs": docs_string,
                }
        endpoints = {}
        if issubclass(clas, PyotCoreBase):
            endpoints = clas.Meta.rules
        extends = []
        for cl in clas.__mro__[1:]:
            cl_name = PyotDocTypeSerializer(cl, []).data
            if cl_name.endswith(("base.PyotCore", "base.PyotStatic", "PyotUtilBase")):
                break
            extends.append(cl_name)
        return {
            "extends": extends,
            "endpoints": endpoints,
            "query_params": query_params,
            "body_params": body_params,
            "definitions": dict(filter(lambda item: item[1]["type"].endswith("definition"), o.items())),
            "attributes": dict(filter(lambda item: item[1]["type"].endswith("attribute"), o.items())),
            "properties": dict(filter(lambda item: item[1]["type"].endswith(("property", "lazy_property")), o.items())),
            "methods": dict(filter(lambda item: item[1]["type"].endswith("method"), o.items())),
        }


class UtilsDocEngine(DocEngine):

    BASE_PATH = Path.cwd() / 'docs' / 'utils'

    def prepare(self):
        '''Activate all models and remove existing docs'''
        shutil.rmtree(self.BASE_PATH, ignore_errors=True)

    def build(self):
        from pyot import utils
        (self.BASE_PATH).mkdir(parents=True, exist_ok=True)
        with open(self.BASE_PATH / 'README.md', 'w+') as f:
            f.writelines(["# Utils"])
        self.build_by_path(utils.__path__)

    def build_by_path(self, path: Iterable[str], nested_levels=tuple()):
        for module in pkgutil.iter_modules(path):
            module = import_module(f"pyot.utils{'.' if nested_levels else ''}{'.'.join(nested_levels)}.{module.name}")
            members = {}
            jump_next_module = False
            for key, obj in inspect.getmembers(module):
                if inspect.ismodule(obj) and obj.__name__.startswith("pyot.utils"):
                    middle_module = obj.__name__.split(".")[-2]
                    (self.BASE_PATH / middle_module).mkdir(parents=True, exist_ok=True)
                    with open(self.BASE_PATH / middle_module / 'README.md', 'w+') as f:
                        f.writelines([f"# {middle_module.capitalize()}"])
                    self.build_by_path([path[0] + '\\' + middle_module], (*nested_levels, middle_module))
                    jump_next_module = True
                    break
                if key.startswith("__") or inspect.ismodule(obj) or isinstance(obj, TypeVar) or (not isinstance(obj, (str, bool, int, float, list)) and obj.__module__ != module.__name__):
                    continue
                if inspect.isclass(obj):
                    is_alias = obj.__name__ != key
                    if not is_alias:
                        members[key] = {"type": "class", "docs": obj.__doc__, **self.get_class_details(obj)}
                    else:
                        members[key] = {"type": "alias", "to": obj.__name__}
                elif inspect.isfunction(obj):
                    is_alias = obj.__name__ != key
                    if not is_alias:
                        members[key] = {"type": "function", "docs": obj.__doc__, **self.get_func_details(obj)}
                    else:
                        members[key] = {"type": "alias", "to": obj.__name__}
                else:
                    members[key] = {"type": "constant", "value": str(obj)}
            
            if jump_next_module:
                continue

            module_lastname = module.__name__.split(".")[-1]
            open_path = self.BASE_PATH
            for nested_level in nested_levels:
                open_path /= nested_level
            with open(open_path / (module_lastname + '.md'), 'w+') as f:
                lines = [
                    f'# {module_lastname.capitalize()} \n\n',
                    f'Module: `{module.__name__}` \n\n',
                ]
                for key, member_info in members.items():
                    if member_info["type"] == "class":
                        lines.append(f'### _class_ `{key}`' + '\n')
                        if member_info['docs']:
                            doc_str = '\n> '.join(inspect.cleandoc(member_info['docs']).split('\n'))
                            lines.append(f"\n> {doc_str}\n")
                        if member_info['extends']:
                            lines.append(f'\nExtends: \n')
                            lines.extend([f'* `{ext}` \n' for ext in member_info['extends']])
                        if member_info["definitions"]:
                            lines.append(f'\nDefinitions: \n')
                            for def_name, definition in member_info['definitions'].items():
                                lines.extend([
                                    f"* `{def_name}` -> `{definition['returns']}` \n",
                                    *[f"  * `{arg_name}`: `{arg_type}` \n" for arg_name, arg_type in definition['args'].items()]
                                ])
                        if member_info['methods']:
                            lines.append(f'\nMethods: \n')
                            for att_name, att_content in member_info['methods'].items():
                                lines.append(f"* _{att_content['type']}_ `{att_name}` -> `{att_content['returns']}` \n")
                                if att_content['args']:
                                    lines.extend([f"  * `{arg_name}`: `{arg_type}` \n" for arg_name, arg_type in att_content['args'].items()])
                                if att_content['docs']:
                                    doc_str = '\n  > '.join(inspect.cleandoc(att_content['docs']).split('\n'))
                                    lines.append(f"  > {doc_str} \n")
                        if member_info['attributes']:
                            lines.append(f'\nAttributes: \n')
                            for att_name, att_content in member_info['attributes'].items():
                                lines.append(f"* `{att_name}` -> `{att_content['returns']}` \n")
                        if member_info['properties']:
                            lines.append(f'\nProperties: \n')
                            for att_name, att_content in member_info['properties'].items():
                                lines.append(f"* `{att_name}` -> `{att_content['returns']}` \n")
                    elif member_info["type"].endswith("function"):
                        lines.append(f"### _{member_info['type']}_ `{key}` -> `{member_info['returns']}` \n")
                        if member_info['args']:
                            lines.extend([f"* `{arg_name}`: `{arg_type}` \n" for arg_name, arg_type in member_info['args'].items()])
                        if member_info['docs']:
                            doc_str = '\n> '.join(inspect.cleandoc(member_info['docs']).split('\n'))
                            lines.append(f"> {doc_str} \n")
                    elif member_info["type"] == "alias":
                        lines.append(f"### _{member_info['type']}_ `{key}` ~ `{member_info['to']}` \n")
                    elif member_info["type"] == "constant":
                        lines.append(f"### _{member_info['type']}_ `{key}`: `{member_info['value']}` \n")
                    lines.append('\n\n')
                f.writelines(lines)

    def get_func_details(self, func: Callable, is_method=False, member_class=None):
        key = func.__name__.split(".")[-1]
        member_type = "method" if is_method else "function"
        args = None
        docs_string = None
        if key.startswith("__") and key.endswith("__"):
            member_type = "definition"
        elif key.startswith("_"):
            return
        if isinstance(func, property):
            member_type = "property"
            return_type = inspect.signature(func.fget).return_annotation
            docs_string = func.fget.__doc__
        else:
            if member_type == "method" and member_class and func.__name__ in member_class.__dict__:
                if isinstance(member_class.__dict__[func.__name__], classmethod):
                    member_type = "classmethod"
                elif isinstance(member_class.__dict__[func.__name__], staticmethod):
                    member_type = "staticmethod"
            return_type = inspect.signature(func).return_annotation
            docs_string = func.__doc__
            args = {}
            for paramname, typ in inspect.signature(func).parameters.items():
                if paramname == "self":
                    continue
                typ_anno = typ.annotation
                if typ.annotation == inspect._empty:
                    if typ.default != inspect._empty and typ.default is not None:
                        typ_anno = typ.default.__class__
                    else:
                        typ_anno = None
                args[paramname] = PyotDocTypeSerializer(typ_anno, []).data
                if args[paramname] is not None and typ.default != inspect._empty:
                    args[paramname] += " = " + str(typ.default)
        if return_type is inspect._empty:
            return_type = None
        if key == "__str__":
            return_type = "str"
        if key == "__len__":
            return_type = "int"
        if inspect.iscoroutinefunction(func):
            member_type = "async" + member_type
        return {
            "args": args,
            "returns": PyotDocTypeSerializer(return_type, []).data,
            "type": member_type,
            "docs": docs_string,
        }

    def get_class_details(self, clas):
        o = {}
        for key, typ in get_type_hints(clas).items():
            if key[0] == "_" and key[-1] != "_":
                continue
            o[key] = {
                "args": None,
                "returns": PyotDocTypeSerializer(typ, []).data,
                "type": "attribute",
                "docs": None,
            }
        for key, func in get_method_properties(clas):
            if key not in o and (not key.startswith("_") or key.startswith("__")) :
                func_details = self.get_func_details(func, True, clas)
                if func_details:
                    o[key] = func_details
        extends = []
        for cl in clas.__mro__[1:]:
            cl_name = PyotDocTypeSerializer(cl, []).data
            if cl_name.endswith(("object")):
                break
            extends.append(cl_name)
        return {
            "extends": extends,
            "definitions": dict(filter(lambda item: item[1]["type"].endswith("definition"), o.items())),
            "attributes": dict(filter(lambda item: item[1]["type"].endswith("attribute"), o.items())),
            "properties": dict(filter(lambda item: item[1]["type"].endswith(("property", "lazy_property")), o.items())),
            "methods": dict(filter(lambda item: item[1]["type"].endswith("method"), o.items())),
        }


class ExamplesDocEngine(DocEngine):

    
    BASE_PATH = Path.cwd() / 'docs' / 'examples'

    def prepare(self):
        '''Activate all models and remove existing docs'''
        shutil.rmtree(self.BASE_PATH, ignore_errors=True)

    def build(self):
        (self.BASE_PATH).mkdir(parents=True, exist_ok=True)
        with open(self.BASE_PATH / 'README.md', 'w+') as f:
            f.writelines(["# Examples"])
        self.build_by_path("examples")

    def build_by_path(self, path: str):
        for dirpath, dirnames, filenames in os.walk(path):
            if len(dirpath.split("\\")) != 2:
                continue
            with open(dirpath + "/docs.md") as rf, open(self.BASE_PATH / (dirpath.split("\\")[-1] + ".md"), "w") as wf:
                for line in rf.readlines():
                    if line.startswith("!!!File="):
                        filepath = line.replace("!!!File=", "").strip()
                        wf.writelines([
                            f"File: `{filepath}`\n",
                            "```python\n",
                            (Path(dirpath) / filepath).read_text() + "\n",
                            "```\n\n"
                        ])
                    else:
                        wf.write(line)




from datetime import datetime, timedelta
from typing import List, Iterator, TYPE_CHECKING

from pyot.conf.model import models
from pyot.core.functional import empty
from .base import PyotCore, PyotStatic

if TYPE_CHECKING:
    from ..riot.account import Account


# PYOT STATIC OBJECTS

class MatchInfoData(PyotStatic):
    id: str
    map_url: str
    start_millis: int
    length_millis: int
    provisioning_flow_id: str
    is_completed: bool
    custom_game_name: str
    queue_id: str
    game_mode: str
    game_version: str
    is_ranked: bool
    season_id: str

    class Meta(PyotStatic.Meta):
        renamed = {"game_length_millis": "length_millis", "game_start_millis": "start_millis", "match_id": "id", "map_id": "map_url"}

    @property
    def start(self) -> datetime:
        return datetime.fromtimestamp(self.start_millis//1000)

    @property
    def length(self) -> timedelta:
        return timedelta(seconds=self.length_millis)


class MatchPlayerAbilityCastData(PyotStatic):
    grenade_casts: int
    ability1_casts: int
    ability2_casts: int
    ultimate_casts: int


class MatchPlayerAbilityData(PyotStatic):
    grenade_effects: int
    ability1_effects: int
    ability2_effects: int
    ultimate_effects: int


class MatchPlayerStatData(PyotStatic):
    score: int
    rounds_played: int
    kills: int
    deaths: int
    assists: int
    playtime_millis: int
    ability_casts: MatchPlayerAbilityCastData

    @property
    def playtime(self) -> timedelta:
        return timedelta(milliseconds=self.playtime_millis)


class MatchPlayerData(PyotStatic):
    puuid: str
    game_name: str
    tag_line: str
    team_id: str
    party_id: str
    character_id: str
    stats: MatchPlayerStatData
    competitive_tier: int
    player_card_id: str
    player_title_id: str

    class Meta(PyotStatic.Meta):
        renamed = {"player_card": "player_card_id", "player_title": "player_title_id"}

    @property
    def account(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)


class MatchTeamData(PyotStatic):
    id: str
    won: bool
    rounds_played: int
    rounds_won: int
    num_points: int

    class Meta(PyotStatic.Meta):
        renamed = {"team_id": "id"}


class MatchLocationData(PyotStatic):
    x: int
    y: int


class MatchPlayerLocationData(PyotStatic):
    puuid: str
    view_radians: float
    location: MatchLocationData


class MatchPlayerFinishingDamageData(PyotStatic):
    damage_type: str
    damage_item: str
    is_secondary_fire_mode: bool


class MatchPlayerKillData(PyotStatic):
    game_time_millis: int
    round_time_millis: int
    killer_puuid: str
    victim_puuid: str
    victim_location: MatchLocationData
    assistant_puuids: List[str]
    player_locations: List[MatchPlayerLocationData]
    finishing_damage: MatchPlayerFinishingDamageData

    class Meta(PyotStatic.Meta):
        raws = {"assistant_puuids"}
        renamed = {"killer": "killer_puuid", "victim": "victim_puuid", "assistants": "assistant_puuids",
            "time_since_game_start_millis": "game_time_millis", "time_since_round_start_millis": "round_time_millis"}

    @property
    def game_time(self) -> timedelta:
        return timedelta(milliseconds=self.game_time_millis)

    @property
    def round_time(self) -> timedelta:
        return timedelta(milliseconds=self.round_time_millis)

    @property
    def killer(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.killer_puuid).using(self.metapipeline.name)

    @property
    def victim(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.victim_puuid).using(self.metapipeline.name)

    @property
    def assistants(self) -> List["Account"]:
        from ..riot.account import Account
        return [Account(puuid=i).using(self.metapipeline.name) for i in self.assistant_puuids]


class MatchPlayerDamageData(PyotStatic):
    receiver: str
    damage: int
    legshots: int
    bodyshots: int
    headshots: int


class MatchPlayerEconomyData(PyotStatic):
    loadout_value: int
    weapon_id: str
    armor_id: str
    remaining: int
    spent: int

    class Meta(PyotStatic.Meta):
        renamed = {"armor": "armor_id", "weapon": "weapon_id"}


class MatchPlayerRoundStatData(PyotStatic):
    puuid: str
    kills: List[MatchPlayerKillData]
    damage: List[MatchPlayerDamageData]
    score: int
    economy: MatchPlayerEconomyData
    ability: MatchPlayerAbilityData


class MatchRoundResultData(PyotStatic):
    round_num: int
    round_result: str
    round_ceremony: str
    winning_team: str
    bomb_planter_puuid: str
    bomb_defuser_puuid: str
    plant_round_millis: int
    plant_player_locations: List[MatchPlayerLocationData]
    plant_location: MatchLocationData
    plant_site: str
    defuse_round_millis: int
    defuse_player_locations: List[MatchPlayerLocationData]
    defuse_location: MatchLocationData
    player_stats: List[MatchPlayerRoundStatData]
    round_result_code: str

    class Meta(PyotStatic.Meta):
        renamed = {"bomb_planter": "bomb_planter_puuid", "bomb_defuser": "bomb_defuser_puuid",
            "plant_round_time": "plant_round_millis", "defuse_round_time": "defuse_round_millis"}

    @property
    def plant_round_time(self) -> timedelta:
        return timedelta(milliseconds=self.plant_round_millis)

    @property
    def defuse_round_time(self) -> timedelta:
        return timedelta(milliseconds=self.defuse_round_millis)

    @property
    def bomb_planter(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.bomb_planter_puuid).using(self.metapipeline.name)

    @property
    def bomb_defuser(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.bomb_defuser_puuid).using(self.metapipeline.name)


class MatchCoachData(PyotStatic):
    puuid: str
    team_id: str


# PYOT CORE OBJECTS

class Match(PyotCore):
    id: str
    info: MatchInfoData
    players: List[MatchPlayerData]
    teams: List[MatchTeamData]
    coaches: List[MatchCoachData]
    round_results: List[MatchRoundResultData]
    # <~> MatchHistory
    start_time_millis: int
    team_id: str
    queue_id: str

    class Meta(PyotCore.Meta):
        renamed = {"match_info": "info", "game_start_time_millis": "start_time_millis", "match_id": "id"}
        rules = {"match_v1_match": ["id"]}

    def __init__(self, id: str = empty, platform: str = models.val.DEFAULT_PLATFORM):
        self.initialize(locals())

    @property
    def start_time(self) -> datetime:
        return datetime.fromtimestamp(self.start_time_millis//1000)


class MatchHistory(PyotCore):
    puuid: str
    history: List[Match]

    class Meta(PyotCore.Meta):
        rules = {"match_v1_matchlist": ["puuid"]}

    def __init__(self, puuid: str = empty, platform: str = models.val.DEFAULT_PLATFORM):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.history[item]

    def __iter__(self) -> Iterator[Match]:
        return iter(self.history)

    def __len__(self):
        return len(self.history)

    @property
    def account(self) -> "Account":
        from ..riot.account import Account
        return Account(puuid=self.puuid).using(self.metapipeline.name)


class RecentMatches(PyotCore):
    current_timestamp: int
    match_ids: List[str]

    class Meta(PyotCore.Meta):
        raws = {"match_ids"}
        rules = {"match_v1_recent": ["queue"]}
        renamed = {"current_time": "current_timestamp"}

    def __init__(self, queue: str = empty, platform: str = models.val.DEFAULT_PLATFORM):
        self.initialize(locals())

    def __getitem__(self, item):
        if not isinstance(item, int):
            return super().__getitem__(item)
        return self.matches[item]

    def __iter__(self) -> Iterator[Match]:
        return iter(self.matches)

    def __len__(self):
        return len(self.matches)

    @property
    def current_time(self) -> datetime:
        return datetime.fromtimestamp(self.current_timestamp)

    @property
    def matches(self)  -> List[Match]:
        return [Match(id=id_, platform=self.platform) for id_ in self.match_ids]



from typing import List, TYPE_CHECKING

from pyot.core.functional import empty
from .base import PyotStatic, PyotCore

if TYPE_CHECKING:
    from .item import Item
    from .merakichampion import MerakiChampion


# PYOT STATIC OBJECTS

class MerakiItemStatDetailData(PyotStatic):
    flat: float
    percent: float
    per_level: float
    percent_per_level: float
    percent_base: float
    percent_bonus: float


class MerakiItemStatData(PyotStatic):
    ability_power: MerakiItemStatDetailData
    armor: MerakiItemStatDetailData
    armor_penetration: MerakiItemStatDetailData
    attack_damage: MerakiItemStatDetailData
    attack_speed: MerakiItemStatDetailData
    cooldown_reduction: MerakiItemStatDetailData
    critical_strike_chance: MerakiItemStatDetailData
    gold_per_10: MerakiItemStatDetailData
    heal_and_shield_power: MerakiItemStatDetailData
    health: MerakiItemStatDetailData
    health_regen: MerakiItemStatDetailData
    lethality: MerakiItemStatDetailData
    lifesteal: MerakiItemStatDetailData
    tenacity: MerakiItemStatDetailData
    magic_penetration: MerakiItemStatDetailData
    magic_resistance: MerakiItemStatDetailData
    mana: MerakiItemStatDetailData
    mana_regen: MerakiItemStatDetailData
    movespeed: MerakiItemStatDetailData
    ability_haste: MerakiItemStatDetailData
    omnivamp: MerakiItemStatDetailData


class MerakiItemPassiveData(PyotStatic):
    unique: bool
    name: str
    effects: str
    range: int
    stats: MerakiItemStatData


class MerakiItemActiveData(PyotStatic):
    unique: bool
    name: str
    effects: str
    range: int
    cooldown: int


class MerakiItemShopPriceData(PyotStatic):
    total: int
    combined: int
    sell: int


class MerakiItemShopData(PyotStatic):
    prices: MerakiItemShopPriceData
    purchasable: bool
    tags: List[str]

    class Meta(PyotStatic.Meta):
        raws = {"tags"}


# PYOT CORE OBJECTS

class MerakiItem(PyotCore):
    name: str
    id: int
    tier: int
    rank: List[str]
    builds_from_ids: List[int]
    builds_into_ids: List[int]
    no_effects: bool
    removed: bool
    required_champion_key: str
    required_ally: str
    icon: str
    simple_description: str
    icon_overlay: bool
    special_recipe_id: int
    nicknames: List[str]
    passives: List[MerakiItemPassiveData]
    active: List[MerakiItemActiveData]
    stats: MerakiItemStatData
    shop: MerakiItemShopData

    class Meta(PyotCore.Meta):
        server_type = "locale"
        rules = {"meraki_item_by_id": ["id"]}
        raws = {"builds_from_ids", "builds_into_ids", "nicknames", "rank"}
        renamed = {"builds_from": "builds_from_ids", "builds_into": "builds_into_ids",
            "required_champion": "required_champion_key", "special_recipe": "special_recipe_id"}

    def __init__(self, id: int = empty):
        self.initialize({"locale": "default", **locals()})

    @property
    def item(self) -> "Item":
        from .item import Item
        return Item(id=self.id, locale="en_us")

    @property
    def builds_from(self) -> List["MerakiItem"]:
        mutable = []
        for i in self.builds_from_ids:
            mutable.append(MerakiItem(id=i))
        return mutable

    @property
    def builds_into(self) -> List["MerakiItem"]:
        mutable = []
        for i in self.builds_into_ids:
            mutable.append(MerakiItem(id=i))
        return mutable

    @property
    def required_champion(self) -> "MerakiChampion":
        from .merakichampion import MerakiChampion
        return MerakiChampion(id=self.required_champion_key)

    @property
    def special_recipe(self) -> "MerakiItem":
        return MerakiItem(id=self.special_recipe_id)



from typing import Any


class UntypedAttribute(Exception):
    '''Exception for flagging an untyped attribute in pyot objects'''

    def __init__(self, clas, key: str, value: Any) -> None:
        super().__init__(f"Untyped attribute '{clas.__name__}.{key}': {value}")



from typing import Mapping


class PyotException(Exception):
    code: int


class NotFindable(PyotException):

    code = 600

    def __init__(self):
        PyotException.__init__(self, "[600] Pipeline token matching pair not found")


class SessionNotFound(PyotException):

    code = 601

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[601] Session Not Found. Origin: {origin}")


class DecodeError(PyotException):

    code = 602

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[602] AioHttp took too long to decode the response. Origin: {origin}")


class NoContent(PyotException):

    code = 204

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[204] No Content. Origin: {origin}")


class NotFound(PyotException):

    code = 404

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[404] Data Not Found. Origin: {origin}")


class MethodNotAllowed(PyotException):

    code = 405

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[405] Method Not Allowed. Origin: {origin}")


class ServerError(PyotException):
    messages: Mapping[int, str] = {
        500: "Internal Server Error",
        502: "Bad Gateway",
        503: "Service Unavailable",
        504: "Gateway Timeout",
    }

    def __init__(self, code, origin="Unknown"):
        self.code = code
        PyotException.__init__(self, f"[{code}] {self.messages[code]}. Origin: {origin}")


class RateLimited(PyotException):

    code = 429

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[429] Rate limit Exceeded. Origin: {origin}")


class Forbidden(PyotException):

    code = 403

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[403] Access Forbidden. Origin: {origin}")


class Unauthorized(PyotException):

    code = 401

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[401] Access Unauthorized. Origin: {origin}")


class BadRequest(PyotException):

    code = 400

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[400] Bad Request. Origin: {origin}")


class Timeout(PyotException):

    code = 408

    def __init__(self, origin="Unknown"):
        PyotException.__init__(self, f"[408] Timeout Connection. Origin: {origin}")


class UnidentifiedResponse(PyotException):

    def __init__(self, code, origin="Unknown"):
        self.code = code
        PyotException.__init__(self, f"[{code}] Unidentified Response {code}. Origin: {origin}")



from typing import Any
import aioredis
import pickle

from pyot.core.exceptions import NotFound
from pyot.pipeline.token import PipelineToken
from pyot.pipeline.expiration import ExpirationManager
from pyot.core.resources import ResourceTemplate
from pyot.utils.logging import LazyLogger

from .base import Store, StoreType


LOGGER = LazyLogger(__name__)


class RedisCache(Store):

    type = StoreType.CACHE

    def __init__(self, game: str, host='127.0.0.1', port=6379, db=0, expirations: Any = None, log_level: int = 0, **kwargs) -> None:
        self.game = game
        self.location = f"{host}:{port}/{db}"
        self.redises = ResourceTemplate(
            acquire_func=lambda: aioredis.Redis(host=host, port=port, db=db, **kwargs),
            release_func=lambda x: x.close(),
        )
        self.expirations = ExpirationManager(game, expirations)
        self.log_level = log_level

    async def get(self, token: PipelineToken, **kwargs) -> Any:
        timeout = self.expirations.get_timeout(token.method)
        if timeout == 0:
            raise NotFound(token.value)
        cache = await self.redises.acquire()
        item = await cache.get(token.value)
        if item is None:
            raise NotFound(token.value)
        LOGGER.log(self.log_level, f"[pyot.stores.rediscache:RedisCache#{self.location}] GET {token.value}")
        return pickle.loads(item)

    async def set(self, token: PipelineToken, value: Any, **kwargs) -> None:
        timeout = self.expirations.get_timeout(token.method)
        if timeout != 0:
            if timeout == -1:
                timeout = None
            cache = await self.redises.acquire()
            await cache.set(token.value, pickle.dumps(value), ex=timeout)
            LOGGER.log(self.log_level, f"[pyot.stores.rediscache:RedisCache#{self.location}] SET {token.value}")

    async def delete(self, token: PipelineToken, **kwargs) -> None:
        cache = await self.redises.acquire()
        await cache.delete(token.value)
        LOGGER.log(self.log_level, f"[pyot.stores.rediscache:RedisCache#{self.location}] DELETE {token.value}")

    async def contains(self, token: PipelineToken, **kwargs) -> bool:
        cache = await self.redises.acquire()
        item = await cache.get(token.value)
        if item is None:
            return False
        return True

    async def clear(self, **kwargs):
        cache = await self.redises.acquire()
        await cache.flushdb()
        LOGGER.log(self.log_level, f"[pyot.stores.rediscache:RedisCache#{self.location}] Store cleared successfully")



from time import time
from typing import Dict, List, Union

from aiohttp import ClientResponse
import aioredis

from pyot.core.resources import ResourceTemplate

from .base import BaseLimiter, LimiterToken


class RedisLimiter(BaseLimiter):

    def __init__(self, game: str, api_key: str, host='127.0.0.1', port=6379, db=0, limiting_share=1, **kwargs):
        self.game = game
        self.api_key = api_key
        self.api_hash = api_key[-8:]
        self.limiting_share = limiting_share
        self.redises = ResourceTemplate(
            acquire_func=lambda: aioredis.Redis(host=host, port=port, db=db, **kwargs),
            release_func=lambda redis: redis.close(),
        )
        self.shas: Dict[str, str] = None
        self.latency = 0.002

    async def register_scripts(self):
        get_token_lua = '''
        local sleep = 0
        local allowed = {}
        local pinging_list = {}
        local app_prefix = KEYS[1]
        local method_prefix = KEYS[2]
        local now = tonumber(ARGV[1])
        local prefix_groups = {{app_prefix, 0, "app"}, {app_prefix, 1, "app"}, {method_prefix, 0, "method"}, {method_prefix, 1, "method"}}
        for j=1, 4 do
            local outerbreak = 0
            repeat
                local prefix, i, type = unpack(prefix_groups[j])
                local prefix_i = prefix..":"..i
                local exists = redis.call("GET", prefix_i..":exists")
                if exists == "0" then
                    break
                end
                local freeze = tonumber(redis.call("GET", prefix_i..":freeze"))
                if (freeze or 0) > now then
                    sleep = math.max(sleep, freeze - now)
                    break
                end
                local pinging = redis.call("GET", prefix_i..":pinging")
                local pingexp = tonumber(redis.call("GET", prefix_i..":pingexp"))
                if now > (pingexp or 0) then
                    pinging = 0
                end
                if pinging == "1" then
                    sleep = math.max(sleep, 0.1)
                    outerbreak = 1
                    break
                end
                local maxcall = tonumber(redis.call("GET", prefix_i..":maxcall"))
                local rollover = tonumber(redis.call("GET", prefix_i..":rollover"))
                local timespan = tonumber(redis.call("GET", prefix_i..":timespan"))
                local called = tonumber(redis.call("GET", prefix_i..":called"))
                local begintime = tonumber(redis.call("GET", prefix_i..":begintime"))
                local pingtime = tonumber(redis.call("GET", prefix_i..":pingtime"))
                local blackout
                local nextstart
                local qualified
                if maxcall ~= nil then
                    blackout = begintime + timespan
                    nextstart = blackout + pingtime
                    qualified = nextstart - now
                end
                if maxcall == nil or now > nextstart then
                    redis.call("SET", prefix_i..":pinging", 1)
                    redis.call("SET", prefix_i..":pingexp", tostring(now + 10))
                    redis.call("SET", prefix_i..":pingbegintime", tostring(now))
                    table.insert(pinging_list, {prefix_i, type, i})
                    break
                end
                if (rollover or 0) + called >= maxcall then
                    sleep = math.max(sleep, qualified)
                    break
                end
                if now >= blackout then
                    sleep = math.max(sleep, qualified)
                    break
                end
                table.insert(allowed, prefix_i)
                break
            until 1
            if outerbreak == 1 then
                break
            end
        end
        if sleep == 0 then
            for i=1, #allowed do
                local allow = allowed[i]
                redis.call("INCR", allow..":called")
                if #pinging_list == 0 then
                    redis.call("INCR", allow..":flying")
                end
            end
        else
            for i=1, #pinging_list do
                local prefix_i, type = unpack(pinging_list[i])
                redis.call("SET", prefix_i..":pinging", 0)
            end
            pinging_list = {}
        end
        local pinging_string = ""
        for i=1, #pinging_list do
            for j=1, #pinging_list[i] do
                pinging_string = pinging_string..tostring(pinging_list[i][j])..","
            end
            pinging_string = pinging_string..";"
        end
        local allowed_string = ""
        for i=1, #allowed do
            allowed_string = allowed_string..tostring(allowed[i])..";"
        end
        return {tostring(sleep), allowed_string, pinging_string}
        '''
        sync_rates_ping_lua = '''
        local now = tonumber(ARGV[1])
        for i=1, #KEYS do
            local keyi = i * 4 - 3
            local prefix_i = KEYS[i]
            local exists = ARGV[keyi + 1]
            redis.call("SET", prefix_i..":exists", exists)
            if exists == "1" then
                redis.call("SET", prefix_i..":maxcall", ARGV[keyi + 2])
                redis.call("SET", prefix_i..":timespan", ARGV[keyi + 3])
                redis.call("SET", prefix_i..":called", ARGV[keyi + 4])
                local flying = redis.call("GET", prefix_i..":flying")
                redis.call("SET", prefix_i..":rollover", (flying or 0))
                redis.call("SET", prefix_i..":flying", 0)
                local begintime = tonumber(redis.call("GET", prefix_i..":pingbegintime"))
                redis.call("SET", prefix_i..":begintime", tostring(begintime))
                redis.call("SET", prefix_i..":pingtime", tostring(now - begintime))
                redis.call("SET", prefix_i..":pinging", 0)
            end
        end
        return 1
        '''
        sync_rates_land_lua = '''
        local epoch = tonumber(ARGV[1])
        for i=1, #KEYS do
            local prefix_i = KEYS[i]
            local begintime = tonumber(redis.call("GET", prefix_i..":begintime"))
            if epoch >= begintime then
                redis.call("DECR", prefix_i..":flying")
            end
        end
        return 1
        '''
        ping_fail_lua = '''
        for i=1, #KEYS do
            redis.call("SET", KEYS[i]..":pinging", 0)
        end
        '''
        freeze_rates_lua = '''
        local now = tonumber(ARGV[1])
        local time = tonumber(ARGV[2])
        for i=1, #KEYS do
            redis.call("SET", KEYS[i]..":freeze", tostring(now + time))
        end
        '''
        redis = await self.redises.acquire()
        return {
            "get_token": await redis.script_load(get_token_lua),
            "sync_rates_ping": await redis.script_load(sync_rates_ping_lua),
            "sync_rates_land": await redis.script_load(sync_rates_land_lua),
            "ping_fail": await redis.script_load(ping_fail_lua),
            "freeze_rates": await redis.script_load(freeze_rates_lua),
        }

    async def get_sha(self, key: str):
        if self.shas is None:
            now = time()
            self.shas = await self.register_scripts()
            self.latency = time() - now
        return self.shas[key]

    async def get_token(self, server: str, method: str):
        app_prefix = f'{self.api_hash}:{self.game}:{server}'
        method_prefix = f'{self.api_hash}:{self.game}:{server}:{method}'
        redis = await self.redises.acquire()
        sha = await self.get_sha("get_token")
        now = time() + self.latency + 0.005
        return_list: List[bytes] = await redis.evalsha(sha, 2, app_prefix, method_prefix, str(now))
        sleep = float(return_list[0].decode('utf-8'))
        allowed = return_list[1].decode('utf-8').split(";")[:-1]
        pinging_list = []
        for pinging in return_list[2].decode('utf-8').split(";")[:-1]:
            pings = pinging.split(",")[:-1]
            pings[-1] = int(pings[-1])
            pinging_list.append(pings)
        return LimiterToken(server, method, now, sleep, allowed, pinging_list)

    async def sync_rates(self, token: LimiterToken, response: ClientResponse) -> Dict[str, List[List[int]]]:
        header = self.parse_headers(response)
        if header is None:
            await self.ping_fail(token)
            return
        redis = await self.redises.acquire()
        keys = []
        now = time() + self.latency + 0.005
        args = [str(now)]
        if token.pinging:
            for (prefix_i, type, i) in token.pinging:
                keys.append(prefix_i)
                if i >= len(header[f'{type}_limit']):
                    for _ in range(4):
                        args.append(0)
                else:
                    args.append(1)
                    args.append(header[f'{type}_limit'][i][0])
                    args.append(header[f'{type}_limit'][i][1])
                    args.append(header[f'{type}_count'][i][0])
            sha = await self.get_sha("sync_rates_ping")
            await redis.evalsha(sha, len(keys), *keys, *args)
        else:
            sha = await self.get_sha("sync_rates_land")
            await redis.evalsha(sha, len(token.allowed), *token.allowed, token.epoch)
        return header

    async def ping_fail(self, token: LimiterToken):
        redis = await self.redises.acquire()
        sha = await self.get_sha("ping_fail")
        keys = [pinging[0] for pinging in token.pinging]
        await redis.evalsha(sha, len(keys), *keys)

    async def freeze_rates(self, token: LimiterToken, response: ClientResponse) -> Dict[str, Union[str, int]]:
        header = self.parse_429(response)
        redis = await self.redises.acquire()
        sha = await self.get_sha("freeze_rates")
        now = time() + self.latency + 0.005
        args = [str(now), str(header["time"])]
        keys = []
        for (prefix_i, type, _) in token.pinging:
            if type != header["type"]:
                continue
            keys.append(prefix_i)
        await redis.evalsha(sha, len(keys), *keys, *args)
        return header



from typing import TypeVar, Generic, List, Iterator

from .copy import fast_copy


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class FrozenGenerator(Generic[T]):
    '''
    Generator that isolates the original list by returning copies of objects when iterated.
    Used for preventing memory leaks of self-filled objects at the cost of performance.
    '''

    def __init__(self, li: List[T]):
        self.objects = li

    def __iter__(self) -> Iterator[T]:
        return (fast_copy(obj) for obj in self.objects)


frozen_generator = FrozenGenerator
